/**********************************************************************
 * $Id: gifsample.cpp,v 1.7 2003/01/31 04:07:55 daniel Exp $
 *
 * Name:     gifsample.cpp
 * Project:  Open Web Toolkit
 * Language: C++
 * Purpose:  Generate GIF sample for chart and map legends.
 * Author:   Daniel Morissette, morissette@dmsolutions.ca 
 *
 **********************************************************************
 * Copyright (c) 1999-2003, DM Solutions Group
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: gifsample.cpp,v $
 * Revision 1.7  2003/01/31 04:07:55  daniel
 * Added support for use as a command-line prg (./gifsample QUERY_STRING=...)
 *
 * Revision 1.6  2002/02/18 16:30:59  julien
 * Bug 812 fix, hatchspacing lesser than 1 don't crash now.
 *
 * Revision 1.5  2002/01/10 18:50:31  daniel
 * Added some clarifications in docs.
 *
 * Revision 1.4  2002/01/10 16:16:30  daniel
 * Call PrintInfo() when PATH_INFO (i.e. image size) is missing
 *
 * Revision 1.3  2002/01/10 15:07:19  julien
 * Added linetype (zigzag, hatch), hatchtype, hatchspacing, new dot symbol,
 * outline, corner
 * Reorganized code to be more expandable
 *
 * Revision 1.2  2000/10/11 21:31:27  assefa
 * Add parameter SymbolId to be able to generate diffrent symbols
 * (triangle, star ...).
 *
 * Revision 1.1  1999/05/12 19:03:19  daniel
 * Initial revision
 *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "owtcgien.h"
#include "cpl_error.h"
#include "cpl_string.h"
#include "gd.h"

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#include <fcntl.h>  /* O_BINARY     */
#endif


#ifdef _WIN32
#   define TMP_ENV   "\\temp\\cgienv.env"
#else
#   define TMP_ENV   "/tmp/cgienv.env"
#endif

int     CreateLineGIFSample(gdImagePtr *im, int nWidth, int nHeight, 
                            int nColor,
                            const char *pszLineStyle, 
                            const char *pszLineWidth,
                            const char *pszLineType,
                            const char *pszHatchType,
                            const char *pszHatchspacing);

int     CreateSymbolGIFSample(gdImagePtr *im, int nWidth, int nHeight, 
                              int nColor,
                              int nSymbolId, int nSymbolSize);

int     DrawBorderGIFSample(gdImagePtr *im, int nOutline, int nCorner, 
                            int nWidth, int nHeigh);

void    PrintInfo(GBool bCGIMode);



extern "C" void MyErrorHandler(CPLErr nErrClass, int nErrno, const char *msg)
{
    fprintf(stdout, "Content-type: text/plain\n\n");
    fprintf(stdout, "ERROR %d: %s\n", nErrno, msg);

    exit(1);
}


/**********************************************************************
 *                       main()
 *
 *
 *  Return a rectangular GIF color sample.
 *
 *  The parameters for the color sample to return are taken from the 
 *  PATH_INFO:
 *
 *  /cgi-bin/gifsample/<width>/<height>/<R>/<G>/<B>
 *
 * /<width>/<height> are optional and may also be passed using the width=
 *              and height= QUERY_STRING parameters.
 *
 * /<R>/<G>/<B> are optional and behave the same way as the Color parameter
 *
 * Optional QUERY_STRING params:
 *
 * The default behavior with only the PATH_INFO parameters above is to 
 * generate a solid color sample.  The following QUERY_STRING parameters
 * are optional can be used to generate other types of symbology:
 *
 * LineStyle - The sample will be a single line, using the specified
 *             pattern.  The pattern is a string in which spaces 
 *             correspond to transparent pixels, and any other character
 *             will draw a pixel of the sample's color.
 *             Passing an empty LineStyle value will draw a solid line.
 *             Also note that spaces are invalid in a URL and have to be
 *             replaced with the '+' character.
 *             eg. "..&LineStyle=xxxx++++&..."
 * LineWidth - Width of the line sample in pixels (default is 1 pixel)
 * LineType  - There is 3 types of line : ZigZag, Hatch, Line.
 *             Default is Line. All the type are influenced by 
 *             LineStyle and LineWidth.
 * HatchType - Type of hatch (Default is 1):
 *             1 : Slash
 *             2 : Backslash
 *             3 : Horizontal
 *             4 : Vertical
 *             5 : XXXXX
 *             6 : +++++
 * HatchSpacing - Space between line in pixels in the Hatch type (default = 4)
 * SymbolID  - Type of symbol (no default): 
 *             1 : square
 *             2 : circle
 *             3 : triangle
 *             4 : plus (+)
 *             5 : star
 *             6 : dot (filled circle)
 * SymbolSize- Size of the symbol called by SymbolID (in pixels).
 *             The biggest symbol that can enter in the gif is the default
 * Color     - Fore color in RGB (FFFFFF) method (Default is black)
 *             if only this parameter is passed it's the background 
 *             color too
 * BgColor   - Background color of the gif.Default is a transparent white.
 * Outline   - The color of the border around the image. These parameter
 *             must be passed to have a border. His width is 1 pixel.
 * Corner    - Remove 1 pixel at each corner of the image. If the browser
 *             don't support transparency the corner color will be the color
 *             passed in parameter (Ex : Corner=FFFFFF will make black corner).
 *
 ********************************************************************** */

int main(int argc, char *argv[]) 
{
    int     nWidth=50, nHeight=50, nR=0, nG=0, nB=0, nColor, nBgColor=0;
    FILE    *fpOut = NULL;
    const char *pszPathInfo, *pszLineStyle, *pszLineWidth, *pszColor;
    const char *pszSymbolId, *pszSymbolSize, *pszLineType;
    const char *pszHatchType, *pszHatchSpacing;
    const char *pszOutline, *pszCorner;
    int    nSymbolId=-1,   nSymbolSize =-1;  
    int    nOutline=0, nCorner=0, nRed=0, nGreen=0, nBlue=0;
    gdImagePtr im;
    GUInt32 n32Color;
    GBool   bCGIMode = TRUE;

    /*-----------------------------------------------------------------
     * Create a OWTCGIEnv object that will be used to handle CGI stuff.
     * When the program exits the object will be destroyed and any 
     * resources it uses will be freed automagically.
     *----------------------------------------------------------------*/
    OWTCGIEnv oCGIEnv;

    /*-----------------------------------------------------------------
     * If -debug switch was passed, then read input from saved CGI env. file
     *----------------------------------------------------------------*/
    // oCGIEnv.DumpToFile(TMP_ENV);
#ifdef DEBUG
    if (argc >= 2 && EQUAL(argv[1], "-debug"))
        oCGIEnv.InitFromFile(TMP_ENV);
    else
#endif
    /*-----------------------------------------------------------------
     * If program does not appear to be running inside a CGI environment
     * then run as a command-line, expecting the filename of a gifsample
     * param. file and of an output file on the command line.
     *----------------------------------------------------------------*/
        if (argc >= 3 || (strlen(oCGIEnv.GetEnv("QUERY_STRING")) == 0 && 
                          strlen(oCGIEnv.GetEnv("REQUEST_METHOD")) == 0))
        {
            if (argc == 3)
            {
                bCGIMode = FALSE;

                if (EQUALN(argv[1],"QUERY_STRING=", 13))
                {
                    /* Got 'gifsample "QUERY_STRING=..."  <output_file>' 
                     */
                    oCGIEnv.SetEnv("QUERY_STRING",   argv[1]+13);
                    oCGIEnv.SetEnv("REQUEST_METHOD", "GET");
                    oCGIEnv.SetEnv("PATH_INFO",      "");

                    oCGIEnv.InitFromQueryString(argv[1]+13);
                }
                else
                {
                    /* Assume arg[1] is a param file 
                     */
                    oCGIEnv.InitFromFile(argv[1]);
                }

                if (!EQUAL(argv[2], "-"))
                {
                    fpOut = fopen(argv[2], "wb");
                    if (fpOut == NULL)
                    {
                        CPLError(CE_Failure, CPLE_FileIO,
                                 "Cannot create output file %s", argv[2]);
                    }
                }
            }
            else
            {
                PrintInfo(FALSE);
                return 1;
            }
        }

    CPLSetErrorHandler(MyErrorHandler);

    pszPathInfo = oCGIEnv.GetEnv("PATH_INFO");

    /* ----------------------------------------------------------
     * Read the PATH_INFO, and extract the values for size and color
     * ---------------------------------------------------------- */
    if (pszPathInfo && strlen(pszPathInfo) > 0)
    {
        char **papszParams;
        int  nParams;

        papszParams = CSLTokenizeStringComplex(pszPathInfo, "/", FALSE, FALSE);
        nParams = CSLCount(papszParams);

        if (nParams > 0) nWidth  = atoi(papszParams[0]);
        if (nParams > 1) nHeight = atoi(papszParams[1]);
        if (nParams > 2) nR = atoi(papszParams[2]);
        if (nParams > 3) nG = atoi(papszParams[3]);
        if (nParams > 4) nB = atoi(papszParams[4]);

        CSLDestroy(papszParams);

        nWidth = MAX(nWidth, 2);
        nHeight = MAX(nHeight, 2);
    }
    else if ((nWidth = atoi(oCGIEnv.GetFormParam("Width"))) > 0 &&
             (nHeight = atoi(oCGIEnv.GetFormParam("Height"))) > 0 )
    {
        nWidth = MAX(nWidth, 2);
        nHeight = MAX(nHeight, 2);
    }
    else
    {
        /* No PATH_INFO, we're either missing the size parameter or
         * we're not running in a CGI environment
         */
        PrintInfo(TRUE);
        return 1;
    }

    /* ----------------------------------------------------------
     * Read Optional (LINE) parameters
     * 
     * HatchTypeidentified different type of symbols
     * - 1 : Slash
     * - 2 : Backslash
     * - 3 : Horizontal
     * - 4 : Vertical
     * - 5 : XXXXX
     * - 6 : +++++
     * ---------------------------------------------------------- */
    pszLineStyle = oCGIEnv.GetFormParam("LineStyle");
    pszLineWidth = oCGIEnv.GetFormParam("LineWidth");
    pszLineType = oCGIEnv.GetFormParam("LineType");
    pszHatchType = oCGIEnv.GetFormParam("HatchType");
    pszHatchSpacing = oCGIEnv.GetFormParam("HatchSpacing");


    /* ----------------------------------------------------------
     * Read Optional Symbol parameters (SymbolId and SymbolSize)
     *
     * SymbolId identified diffrent types of symbols 
     * - 1 : square
     * - 2 : circle
     * - 3 : triangle
     * - 4 : plus (+)
     * - 5 : star
     * - 6 : dot
     * ---------------------------------------------------------- */
    
    nSymbolId = -1;
    pszSymbolId = oCGIEnv.GetFormParam("SymbolId");
    if (strlen(pszSymbolId) > 0)
      nSymbolId = atoi(pszSymbolId);

    nSymbolSize = -1;
    pszSymbolSize = oCGIEnv.GetFormParam("SymbolSize");
    if (strlen(pszSymbolSize) > 0)
      nSymbolSize = atoi(pszSymbolSize);


    /*----------------------------------------------------------------
     *Create Image
     *---------------------------------------------------------------*/

    /* Allocate image in memory */
     im = gdImageCreate(nWidth, nHeight);

    /* ----------------------------------------------------------
     * Background color can optionally be assign through "BgColor" param
     * ---------------------------------------------------------- */
    pszColor = oCGIEnv.GetFormParam("BgColor");
    if(strlen(pszColor) > 0)
    {
        n32Color = strtoul((char*)pszColor, NULL, 16);
        nRed = (n32Color/0x10000) & 0xff;
        nGreen = (n32Color/0x100) & 0xff;
        nBlue = (n32Color) & 0xff;
        
        /* Allocate first color... will become the bg color */
        nBgColor = gdImageColorAllocate(im, nRed, nGreen, nBlue);
    }
    else
    {
        nBgColor = gdImageColorAllocate(im, 255, 255, 255);
        gdImageColorTransparent(im, nBgColor);
        nBgColor = 0;
    }

    /* --------------------------------------------------------------
     * Read optional (Outline) param
     * -------------------------------------------------------------- */
    pszOutline = oCGIEnv.GetFormParam("Outline");
    pszCorner = oCGIEnv.GetFormParam("Corner");
    if(strlen(pszOutline) > 0)
    {
        n32Color = strtoul((char*)pszOutline, NULL, 16);
        nRed = (n32Color/0x10000) & 0xff;
        nGreen = (n32Color/0x100) & 0xff;
        nBlue = (n32Color) & 0xff;

        nOutline = gdImageColorAllocate(im, nRed, nGreen, nBlue);
    }

    if(strlen(pszCorner) > 0)
    {
        n32Color = strtoul((char*)pszCorner, NULL, 16);
        nRed = (n32Color/0x10000) & 0xff;
        nGreen = (n32Color/0x100) & 0xff;
        nBlue = (n32Color) & 0xff;

        nCorner = gdImageColorAllocate(im, nRed, nGreen, nBlue);
        gdImageColorTransparent(im, nCorner);
    }


    /* ----------------------------------------------------------
     * Sample Color can optionally be passed through the "Color" param
     * ---------------------------------------------------------- */
    pszColor = oCGIEnv.GetFormParam("Color");
    if (strlen(pszColor) > 0)
    {
        n32Color = strtoul((char*)pszColor, NULL, 16);
        nR = (n32Color/0x10000) & 0xff;
        nG = (n32Color/0x100) & 0xff;
        nB = (n32Color) & 0xff;
    }

        nColor = gdImageColorAllocate(im, nR, nG, nB);
    /* ----------------------------------------------------------
     * Output the image 
     * (On WIN32 we have to make sure output file is in binary mode.)
     * ---------------------------------------------------------- */
#ifdef WIN32
    if ( fpOut == NULL &&
         _setmode( _fileno( stdout ), O_BINARY ) == -1)
        CPLError(CE_Failure, CPLE_FileIO, 
                      "Cannot set output file mode (%s)",  strerror(errno));
#endif

    /*----------------------------------------------------------------
     *Create the gif
     *---------------------------------------------------------------*/

    if(strlen(pszLineStyle)>0 || strlen(pszLineWidth)>0 || 
                strlen(pszLineType)>0)
    {
        CreateLineGIFSample(&im, nWidth, nHeight, nColor,
                            pszLineStyle, pszLineWidth,
                            pszLineType, pszHatchType, pszHatchSpacing);
    }
    else if (nSymbolId > 0)
    {
        CreateSymbolGIFSample(&im, nWidth, nHeight, nColor,
                              nSymbolId, nSymbolSize);
    }
    else if (nBgColor == 0)
    {
        gdImageFill(im, 1, 1, nColor);
    }
    
    /*Make a border if necessary*/
    if((nOutline>0) || (strlen(pszCorner)>0))
    {
        DrawBorderGIFSample( &im, nOutline, nCorner, nWidth, nHeight);
    }

    /*----------------------------------------------------------------
     *Save in a file
     *---------------------------------------------------------------*/

    /* Output the image */
    if (bCGIMode)
        fprintf(fpOut?fpOut:stdout, "Content-Type: image/gif\n\n");
    gdImageGif(im, fpOut? fpOut:stdout);
    fflush(fpOut? fpOut:stdout);

    if (fpOut)
        fclose(fpOut);

    /* Destroy image in memory */
    gdImageDestroy(im);

    return 0;
}


/**********************************************************************
 *                      CreateLineGIFSample()
 *
 * Produce the GIF color sample image and return it through fpOut.
 **********************************************************************/
int     CreateLineGIFSample(gdImagePtr *im, int nWidth, int nHeight, 
                            int nColor,
                            const char *pszLineStyle,
                            const char *pszLineWidth,
                            const char *pszLineType,
                            const char *pszHatchType,
                            const char *pszHatchSpacing)
{
    int i, j, y, nLineWidth, nLen=0, *panStyle=NULL;
    int y1, x2, nHatchSpacing, nHatchType, nLine, nWidth13 = nWidth/3;

    /* Build line style and draw the line */
    if (EQUAL(pszLineStyle, ""))
        pszLineStyle = "x";
    nLen = strlen(pszLineStyle);
    panStyle = (int*)CPLMalloc(nLen*sizeof(int));
    for(i=0; i<nLen; i++)
    {
        panStyle[i] = (pszLineStyle[i]==' ')?gdTransparent:nColor;
    }

    nLineWidth = atoi(pszLineWidth);
    nLineWidth = MAX(nLineWidth, 1);
    nLineWidth = MIN(nLineWidth, nHeight);

    if(EQUAL(pszLineType, "ZigZag")) /*ZigZag*/
    {
        for(i=0; i<nLineWidth; i++)
        {
            gdImageSetStyle(*im, panStyle, nLen);
            gdImageLine(*im, i, nHeight,nWidth13 + i, 0, gdStyled);
            gdImageLine(*im, nWidth13 + i, 0, nWidth13 * 2 + i, nHeight, gdStyled);
            gdImageLine(*im, nWidth13 * 2 + i, nHeight, nWidth + i, 0, gdStyled);    
        }
    }
    else
    {
        if(EQUAL(pszLineType, "Hatch")) /*Hatch*/
        {
            nHatchSpacing = atoi (pszHatchSpacing);
            if(nHatchSpacing < 1) /*Default value of Hatch Spacing is 4*/
                nHatchSpacing = 4;

            nHatchType = atoi (pszHatchType);
            nLine = (nWidth+nHeight)/nHatchSpacing; /*how much lines*/
            
            /*Default (Slash) or X*/
            if ((nHatchType == 5) || (nHatchType < 2) || (nHatchType > 6))
            {
                y1 = 0;
                x2 = 0;
                for(i=0; i<=nLine; i++)
                {
                    y1 += nHatchSpacing;
                    x2 = y1;
                    for(j=0; j<nLineWidth; j++)
                    {
                        gdImageSetStyle(*im, panStyle, nLen);
                        gdImageLine(*im, 0, y1++, x2++, 0, gdStyled);
                    }
                }
            }
            /*Backslash or X*/
            if((nHatchType == 2) || (nHatchType == 5))
            {
                y1 = nHeight;
                x2 = 0;
                for(i=0; i<=nLine; i++)
                {
                    y1 -= nHatchSpacing;
                    x2 += nHatchSpacing;
                    for(j=0; j<nLineWidth; j++)
                    {
                        gdImageSetStyle(*im, panStyle, nLen);
                        gdImageLine(*im, 0, y1--, x2++, nHeight, gdStyled);
                    }
                }
            }
            /*Horizontal or +*/
            if((nHatchType == 3) || (nHatchType == 6))
            {
                for(y1=nHatchSpacing; y1<=nHeight; y1+=nHatchSpacing)
                {
                    for(j=0; j<nLineWidth; j++)
                    {
                        gdImageSetStyle(*im, panStyle, nLen);
                        gdImageLine(*im, 0, y1, nWidth, y1, gdStyled);
                        y1++;
                    }
                }
            }
            /*Vertical or +*/
            if((nHatchType == 4) || (nHatchType == 6))
            {
                for(x2=nHatchSpacing; x2<=nWidth; x2+=nHatchSpacing)
                {
                    for(j=0; j<nLineWidth; j++)
                    {
                        gdImageSetStyle(*im, panStyle, nLen);
                        gdImageLine(*im, x2, 0, x2, nHeight, gdStyled);
                        x2++;
                    }
                }
            }
        }
        else /*Centered horizontal line*/
        {
            y = nHeight/2 - nLineWidth/2;
            for(i=0; i<nLineWidth; i++, y++)
            {
                gdImageSetStyle(*im, panStyle, nLen);
                gdImageLine(*im, 0, y, nWidth, y, gdStyled);
            }
        }
    }
    return 0;
}


/**********************************************************************
 *                      CreateSymbolGIFSample()
 *
 * Produce the GIF color sample image and return it through fpOut.
 **********************************************************************/
int     CreateSymbolGIFSample(gdImagePtr *im, int nWidth, int nHeight, 
                              int nColor,
                              int nSymbolId, int nSymbolSize)
{
    gdPoint rectangle[3];
    gdPoint star[10];

    if (nSymbolSize < 0 || (nSymbolSize > nWidth &&  nSymbolSize > nHeight))
      nSymbolSize = MIN(nWidth, nHeight);
    
    switch (nSymbolId)
    { 
        case 1: /*rectangle */
          gdImageFilledRectangle(*im, (nWidth/2 - nSymbolSize/2), 
                                 (nHeight/2 - nSymbolSize/2), 
                                 (nWidth/2 + nSymbolSize/2), 
                                 (nHeight/2 + nSymbolSize/2),
                                 nColor);
          break;

        case 2: /*circle */
          gdImageArc(*im, nWidth/2, nHeight/2, nSymbolSize, nSymbolSize,
                     0, 360, nColor);
          break;

        case 3:/*triangle*/
          rectangle[0].x = (nWidth/2) - (nSymbolSize/2);
          rectangle[0].y = (nHeight/2) + (nSymbolSize/2);
          rectangle[1].x = (nWidth/2);
          rectangle[1].y = (nHeight/2) - (nSymbolSize/2);
          rectangle[2].x = (nWidth/2) + (nSymbolSize/2);
          rectangle[2].y = (nHeight/2) + (nSymbolSize/2);

          gdImageFilledPolygon(*im, rectangle, 3, nColor);
          break;

        case 4: /*plus + */
          gdImageLine (*im, (nWidth/2 - nSymbolSize/2), (nHeight/2),
                       (nWidth/2 + nSymbolSize/2), (nHeight/2),
                       nColor);
          gdImageLine (*im, (nWidth/2), (nHeight/2 - nSymbolSize/2), 
                       (nWidth/2), (nHeight/2 + nSymbolSize/2), 
                       nColor);
          break;

        case 5: /*star*/
          
           star[0].x = (nWidth/2) - (nSymbolSize/2);
           star[0].y = (int)(nHeight*0.375);

           star[1].x = (nWidth/2) - (nSymbolSize/4);
           star[1].y = (int)(nHeight*0.375);

           star[2].x = (nWidth/2);
           star[2].y = (nHeight/2) - (nSymbolSize/2);

           star[3].x = (nWidth/2) + (nSymbolSize/4);
           star[3].y = (int)(nHeight*0.375);

           star[4].x = (nWidth/2) + (nSymbolSize/2);
           star[4].y = (int)(nHeight*0.375);
          
           star[5].x = (nWidth/2) + (nSymbolSize/4);
           star[5].y = (nHeight/2);

           star[6].x = (nWidth/2) + (nSymbolSize/2);
           star[6].y = (nHeight/2) + (nSymbolSize/2);
          
           star[7].x = (nWidth/2);
           star[7].y = (nHeight/2) + (nSymbolSize/4);

           star[8].x = (nWidth/2) - (nSymbolSize/2);
           star[8].y = (nHeight/2) + (nSymbolSize/2);
          
           star[9].x = (nWidth/2) - (nSymbolSize/4);
           star[9].y = (nHeight/2);

          gdImageFilledPolygon(*im, star, 10, nColor);
          break;

       case 6: /*dot*/
          gdImageArc(*im, nWidth/2, nHeight/2, nSymbolSize, nSymbolSize,
                     0, 360, nColor);
          gdImageFillToBorder(*im, nWidth/2, nHeight/2, nColor , nColor);
          break;

        default:
          break;
    }

    return 0;
}


/**********************************************************************
 *                 DrawBorderGIFSample()
 *
 * Draw a border of 1 pixel around the image.
 ********************************************************************** */

int DrawBorderGIFSample(gdImagePtr *im, int nOutline, int nCorner,
                        int nWidth, int nHeight)
{
    /*Make a border if necessary*/
    if(nOutline>0)
    {
        gdImageLine(*im, 0, 0,nWidth, 0, nOutline);
        gdImageLine(*im, 0, 0, 0, nHeight, nOutline);
        gdImageLine(*im, 0, nHeight-1, nWidth, nHeight-1, nOutline);
        gdImageLine(*im, nWidth-1, 0, nWidth-1, nHeight, nOutline);
    }
    if(nCorner>0)
    {
        gdImageSetPixel(*im, 0, 0, nCorner);
        gdImageSetPixel(*im, 0, nHeight-1, nCorner);
        gdImageSetPixel(*im, nWidth-1, 0, nCorner);
        gdImageSetPixel(*im, nWidth-1, nHeight-1, nCorner);
    }
    return 0;
}


/**********************************************************************
 *                       PrintInfo()
 *
 * Return Version and Copyright information to stdout.
 ********************************************************************** */

void PrintInfo(GBool bCGIMode)
{
    if (bCGIMode)
        printf("Content-Type: text/plain\n\n");

    printf("OWT GIF Sample\n\n");
    printf("Copyright (c) 1999-2003, DM Solutions Group, http://www.dmsolutions.ca/\n\n");
    printf("Uses the GD library from Thomas Boutell (http://www.boutell.com/)\n");
    printf("\n");
    printf("$Revision: 1.7 $  $Date: 2003/01/31 04:07:55 $ (Built %s %s)\n", __DATE__, __TIME__);

    if (!bCGIMode)
    {
        /* Print command-line args */
                printf( "\n"
        "Usage: gifsample <param_file.txt> <out_file.gif>\n"
        "   or: gifsample <QUERY_STRING=...> <out_file.gif>\n\n"
        "   <param_file.txt>: the name of the text file that contains the\n"
        "                     gifsample parameters\n\n" 
        "   <out_file.gif>: the name of the GIF file to create. Passing\n"
        "                   a value of '-' will send the GIF to stdout.\n\n");
    }

}

