/**********************************************************************
 * $Id: giftaskbar.cpp,v 1.5 2002/10/03 17:46:49 sacha Exp $
 *
 * Name:     giftaskbar.cpp
 * Project:  Open Web Toolkit
 * Language: C++
 * Purpose:  Generate GIF task bars for HMTL-based Gant charts.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: giftaskbar.cpp,v $
 * Revision 1.5  2002/10/03 17:46:49  sacha
 * Fixed a bug when aligning label.
 *
 * Revision 1.4  1999/11/25 16:47:14  daniel
 * Fixed compile warnings
 *
 * Revision 1.3  1999/11/25 16:41:31  daniel
 * Added optional height param for BARs
 *
 * Revision 1.2  1999/11/24 20:06:36  daniel
 * Added LABEL and VGRID params
 *
 * Revision 1.1  1999/11/17 05:06:31  daniel
 * Initial version
 *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "owtcgien.h"
#include "cpl_error.h"
#include "cpl_string.h"
#include "gd.h"
extern "C" {
#include "gdfonts.h"
}

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#include <fcntl.h>  /* O_BINARY     */
#endif


#ifdef _WIN32
#   define TMP_ENV   "\\temp\\cgienv.env"
#else
#   define TMP_ENV   "/tmp/cgienv.env"
#endif

int     DrawTaskBar(FILE *fpOut, OWTCGIEnv *poCGIEnv, int nWidth, int nHeight, 
                    int nR, int nG, int nB);
void    PrintInfo();


#define COLOR_HEX2RGB(strHex, r, g, b)                  \
    if (strHex && strlen(strHex) > 0)                   \
    {                                                   \
        GUInt32 n32Color;                               \
        n32Color = strtoul((char*)strHex, NULL, 16);    \
        r = (n32Color/0x10000) & 0xff;                  \
        g = (n32Color/0x100) & 0xff;                    \
        b = (n32Color) & 0xff;                          \
    }




extern "C" void MyErrorHandler(CPLErr nErrClass, int nErrno, const char *msg)
{
    fprintf(stdout, "Content-type: text/plain\n\n");
    fprintf(stdout, "ERROR %d: %s\n", nErrno, msg);

    exit(1);
}


/**********************************************************************
 *                       main()
 *
 *
 * Return GIF image to be used as a task bar in a Gant Chart.
 *
 * All the parameters for the bar are passed through parameters in the
 * QUERY_STRING.  
 *
 * Here are the GIF image's base parameter:
 *
 *  W           GIF image Width (default = 300)
 *  H           GIF image eight (default = 20)
 *  BG          Image BG Color (default = FFFFFF = white)
 *
 * These 3 parameters can optionally be passed in the PATH_INFO part of
 * the URL, as in:
 *
 *   /cgi-bin/giftaskbar/<W>/<H>/<R>/<G>/<B>?<...QUERY_SRTING...>
 *
 *      The following 2 parameters define the range of time covered by the
 *      GIF image... location of bars and milestone can then be specified
 *      in time units (days, weeks, etc.) within that range.
 *      Note that a margin equal to half the height of the image will be 
 *      included at the beginning and at the end of the time scale so 
 *      that TMIN and TMAX are both included in the image
 *
 *  TMIN        Time at which the image starts, in time units (default = 0)
 *  TMAX        Time at which the image ends, in time units (default = <W>)
 *
 *      The other parameters specify objects to draw in the bar.  Their
 *      position is specified in time units (see TMIN/TMAX above)
 *      and the objects are drawn in the order in which they appear in the
 *      QUERY_STRING
 *  BAR=<t0>;<t1>;<color>[;<h>] Draw a bar going from time <t0> to <t1>
 *                              using specified color (in hex RGB)
 *                              The optional <h> parameter can be used to
 *                              specify the portion of the height of the
 *                              GIF image this bar will cover (range= ]0..1])
 *                              The default is 0.5 and means that the bar
 *                              is centered vertically and its height is 50%
 *                              of the GIF's height.
 *                              A value of 1 means that the bar covers all
 *                              the height of the GIF image.
 *
 *  POINT=<t>;<symbol>;<color>  Draw a point (milestone) at time <t>
 *                              using specified <color>.
 *                              <symbol> can be:  1 = default = large diamond
 *                                                2 = small diamond
 *                                              ... more to come
 *
 *  LABEL=<t>;<text_string>[;<just>[;<color>]]
 *
 *       Draws the <text_string> label at the specified time position.
 *       <just> is the optional justification ("R", "C" or "L") the default
 *       is left.
 *       <color> is optional, the default is black.
 *       For now, there is no control on the font size... text it is always
 *       aproximately 10 pixels high and centered vertically in the image.
 *
 *  VGRID=<t0>;<t_spacing>;<color>
 *
 *       Draws vertical grid lines starting at time <t0>, and repeatedly
 *       every <t_spacing> time units using the specified color.
 *
 ********************************************************************** */

int main() 
{
    int     nWidth=300, nHeight=20, nR=255, nG=255, nB=255;
    const   char *pszPathInfo, *pszValue;
    FILE    *fpOut;
    /*-----------------------------------------------------------------
     * Create a OWTCGIEnv object that will be used to handle CGI stuff.
     * When the program exits the object will be destroyed and any 
     * resources it uses will be freed automagically.
     *----------------------------------------------------------------*/
    OWTCGIEnv oCGIEnv;

    /*-----------------------------------------------------------------
     * Save/Restore saved CGI Env. State for debugging
     *----------------------------------------------------------------*/
    // oCGIEnv.DumpToFile(TMP_ENV);
#ifdef DEBUG
    if (strlen(oCGIEnv.GetEnv("SCRIPT_NAME")) == 0 &&
        strlen(oCGIEnv.GetEnv("PATH_INFO")) == 0 &&
        strlen(oCGIEnv.GetEnv("QUERY_STRING")) == 0 )
        oCGIEnv.InitFromFile(TMP_ENV);
#endif

    CPLSetErrorHandler(MyErrorHandler);

    pszPathInfo = oCGIEnv.GetEnv("PATH_INFO");

    fpOut = stdout;

    /* ----------------------------------------------------------
     * Read the PATH_INFO, and extract the values for size and BG Color
     * ---------------------------------------------------------- */
    if (pszPathInfo && strlen(pszPathInfo) > 0)
    {
        char **papszParams;
        int  nParams;

        papszParams = CSLTokenizeStringComplex(pszPathInfo, "/", FALSE, FALSE);
        nParams = CSLCount(papszParams);

        if (nParams > 0) nWidth  = atoi(papszParams[0]);
        if (nParams > 1) nHeight = atoi(papszParams[1]);
        if (nParams > 2) nR = atoi(papszParams[2]);
        if (nParams > 3) nG = atoi(papszParams[3]);
        if (nParams > 4) nB = atoi(papszParams[4]);

        CSLDestroy(papszParams);

    }

    /* ----------------------------------------------------------
     * Also check for W, H, and BG color in the QUERY_SRTING
     * ---------------------------------------------------------- */
    if (strlen((pszValue=oCGIEnv.GetFormParam("W"))) > 0) 
        nWidth = atoi(pszValue);
    if (strlen((pszValue=oCGIEnv.GetFormParam("H"))) > 0) 
        nHeight = atoi(pszValue);
    if (strlen((pszValue=oCGIEnv.GetFormParam("BG"))) > 0) 
    {
        COLOR_HEX2RGB(pszValue, nR, nG, nB);
    }

    nHeight = MAX(nHeight, 2);
    nWidth = MAX(nWidth, 2*nHeight);

    /* ----------------------------------------------------------
     * On WIN32 we have to make sure output file is in binary mode.
     * ---------------------------------------------------------- */
#ifdef WIN32
    if ( _setmode( _fileno( fpOut ), O_BINARY ) == -1)
        CPLError(CE_Failure, CPLE_FileIO, "OWTChart::SaveSlideGIF(): "
                      "Cannot set output file mode (%s)",  strerror(errno));
#endif

    /* ----------------------------------------------------------
     * Generate the graphical objects and the image 
     * ---------------------------------------------------------- */

    DrawTaskBar(fpOut, &oCGIEnv, nWidth, nHeight, nR, nG, nB);
    
    return 0;
}

/**********************************************************************
 *                      GetColorFromRGB()
 *
 * Return a GD color index for a RGB color.  If color table is full and
 * requested RGB color is not part of it, then the closest match will be
 * returned.
 **********************************************************************/
int GetColorFromRGB(gdImagePtr im, int nR, int nG, int nB)
{
    int nColorIndex;

    /* Check if color already exists */
    nColorIndex = gdImageColorExact(im, nR, nG, nB);
    /* If it isn't already present... */
    if (nColorIndex == (-1)) 
    {
        /* Second best: try to allocate it directly. */
        nColorIndex = gdImageColorAllocate(im, nR, nG, nB);      
        if (nColorIndex == -1)
        {
            /* Out of colors, so find the closest color instead. */
            nColorIndex = gdImageColorClosest(im, nR, nG, nB);
        }
     }
    return nColorIndex;
}

int GetColorFromHex(gdImagePtr im, const char *pszHexColor)
{
    int nR=0, nG=0, nB=0;
    COLOR_HEX2RGB(pszHexColor, nR, nG, nB);

    return GetColorFromRGB(im, nR, nG, nB);
}

/**********************************************************************
 *                      DrawTaskBar()
 *
 * Draw the objects on the image, and return the GIF output through fpOut.
 **********************************************************************/
int DrawTaskBar(FILE *fpOut, OWTCGIEnv *poCGIEnv, int nWidth, int nHeight, 
                    int nR, int nG, int nB)
{
    gdImagePtr im;
    int nColor;
    double dTMin=0, dTMax=nWidth, dTime2PixScale = 1;
    const char *pszTMin, *pszTMax;
    char **papszParams;

    /* Allocate image in memory */
    im = gdImageCreate(nWidth, nHeight);

    /* Allocate first color... will become the bg color */
    nColor = gdImageColorAllocate(im, nR, nG, nB);

    /* ----------------------------------------------------------
     * Fetch the start and end time and setup the scaling, etc.
     * Note that the scale and the image origin in time units have
     * to be adjusted to keep a maring of half the image height
     * at the beginning and at the end of the bar.
     * ---------------------------------------------------------- */
    if (strlen((pszTMin=poCGIEnv->GetFormParam("TMIN"))) > 0 &&
        strlen((pszTMax=poCGIEnv->GetFormParam("TMAX"))) > 0  )
    {
        /* Read values */
        dTMin = atof(pszTMin);
        dTMax = atof(pszTMax);
        /* Force at least 1 unit of time */
        dTMax = MAX(dTMax, dTMin+1);
        /* Calculate scale and adjust origin */
        dTime2PixScale = (nWidth-nHeight)/(dTMax-dTMin);
        dTMin = dTMin - 0.5*nHeight/dTime2PixScale;
    }

#define TIME2PIX(x) (int)(((x)-dTMin)*dTime2PixScale)

    /* ----------------------------------------------------------
     * Loop on all query parameters, drawing objects in the order they
     * appear... skip unknown parameters without complaining
     * ---------------------------------------------------------- */
    papszParams = poCGIEnv->GetAllFormParams();
    for (int iParam=0; papszParams && papszParams[iParam]; iParam++)
    {
        char **papszTok = CSLTokenizeStringComplex(papszParams[iParam], "=;",
                                                   FALSE, TRUE);
        int numTok = CSLCount(papszTok);

        if (numTok >= 4 && EQUAL(papszTok[0], "BAR"))
        {
            /* --------------------------------------------------
             * BAR=<t0>;<t1>;<color>[;<h>]
             * -------------------------------------------------- */
            double dT0 = atof(papszTok[1]);
            double dT1 = atof(papszTok[2]);
            int nColor = GetColorFromHex(im, papszTok[3]);

            /* Height of the bar, in proportion of the GIF's height */
            double dHeight = 0.5;
            if (numTok >= 5)
                dHeight = atof(papszTok[4]);

            gdImageFilledRectangle(im, TIME2PIX(dT0), 
                                   (int)(nHeight*(1.0-dHeight)/2), 
                                   TIME2PIX(dT1), 
                                   (int)(nHeight*(1.0+dHeight)/2), 
                                   nColor);
        }
        else if (numTok == 4 && EQUAL(papszTok[0], "POINT"))
        {
            /* --------------------------------------------------
             * POINT=<t>;<symbol>;<color>
             * -------------------------------------------------- */
            double dT = atof(papszTok[1]);
            int nSymbol = atoi(papszTok[2]);
            int nColor = GetColorFromHex(im, papszTok[3]);
            gdPoint points[10];

            switch(nSymbol)
            {
              case 1:   // Large Diamond - The default
              default:
                points[0].x = TIME2PIX(dT)-nHeight/2;
                points[0].y = nHeight/2;
                points[1].x = TIME2PIX(dT);
                points[1].y = nHeight;
                points[2].x = TIME2PIX(dT)+nHeight/2;
                points[2].y = nHeight/2;
                points[3].x = TIME2PIX(dT);
                points[3].y = 0;
                points[4].x = TIME2PIX(dT)-nHeight/2;
                points[4].y = nHeight/2;

                gdImageFilledPolygon(im, points, 4, nColor);
                break;
              case 2:   // Small Diamond
                points[0].x = TIME2PIX(dT)-nHeight/4;
                points[0].y = nHeight/2;
                points[1].x = TIME2PIX(dT);
                points[1].y = 3*nHeight/4;
                points[2].x = TIME2PIX(dT)+nHeight/4;
                points[2].y = nHeight/2;
                points[3].x = TIME2PIX(dT);
                points[3].y = nHeight/4;
                points[4].x = TIME2PIX(dT)-nHeight/4;
                points[4].y = nHeight/2;

                gdImageFilledPolygon(im, points, 4, nColor);
                break;
              case 3:   // Triangle pointing up


              case 4:   // Tirangle pointing down
                break;
            }

        }
        else if (numTok >= 3 && EQUAL(papszTok[0], "LABEL"))
        {
            /* --------------------------------------------------
             * LABEL=<t>;<text_string>[;<just>[;<color>]]
             * -------------------------------------------------- */
            double dT = atof(papszTok[1]);
            char *pszText = papszTok[2];

            /* For now, always use "SMALL" font */
            gdFontPtr pFont = gdFontSmall;

            /* Text is always centered vertically in the bar. */
            int nY = nHeight/2 - pFont->h / 2;

            /* Look for justification param. */
            int nX = 0;
            if (numTok >= 4 && EQUAL(papszTok[3], "C"))
            {
                /* justification = Center */
                nX = TIME2PIX(dT) - strlen(pszText)*pFont->w/2;
            }
            else if (numTok >= 4 && EQUAL(papszTok[3], "R"))
            {
                /* justification = Right */
                nX = TIME2PIX(dT) + strlen(pszText)*pFont->w;
            }
            else
            {
                /* Default: Left justified */
                nX = TIME2PIX(dT);
            }

            /* Check if color param was passed, default is black */
            int nColor = 0;
            if (numTok >= 5)
                nColor = GetColorFromHex(im, papszTok[4]);
            else
                nColor = GetColorFromRGB(im, 0, 0, 0);

            /* OK, we're ready to draw the text string */
            gdImageString(im, pFont, nX, nY, (unsigned char *)pszText, nColor);
        }
        else if (numTok == 4 && EQUAL(papszTok[0], "VGRID"))
        {
            /* --------------------------------------------------
             * Vertical grid lines...
             *
             * VGRID=<t0>;<t_spacing>;<color>
             * -------------------------------------------------- */
            double dT0 = atof(papszTok[1]);
            double dTSpacing = atof(papszTok[2]);
            int nColor = GetColorFromHex(im, papszTok[3]);

            while(dT0 < dTMax)
            {
                gdImageLine(im, TIME2PIX(dT0), 0, 
                            TIME2PIX(dT0), nHeight, nColor);

                dT0 += dTSpacing;
            }
        }

        CSLDestroy(papszTok);
    }

    /* Output the image */
    fprintf(fpOut, "Content-Type: image/gif\n\n");
    gdImageGif(im, fpOut);

    /* Destroy image in memory */
    gdImageDestroy(im);

    return 0;
}


/**********************************************************************
 *                       PrintInfo()
 *
 * Return Version and Copyright information to stdout.
 ********************************************************************** */

void PrintInfo()
{
    printf("Content-Type: text/plain\n\n");

    printf("OWT GIF Gant Chart Task Bar\n\n");
    printf("Copyright (c) 1999, Daniel Morissette (danmo@videotron.ca)\n\n");
    printf("Uses the GD (gd1.3) library from Thomas Boutell (http://www.boutell.com/)\n");
    printf("\n");
    printf("$Revision: 1.5 $  $Date: 2002/10/03 17:46:49 $ (Built %s %s)\n", __DATE__, __TIME__);
}

