/**********************************************************************
 * $Id: wsgifanim.cpp,v 1.1 1999/10/07 23:05:24 daniel Exp $
 *
 * Name:     wsgifanim.cpp
 * Project:  Open Web Toolkit
 * Language: C++
 * Purpose:  Generate animated GIF using SPANS Web Server slides from 
 *           the cache.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: wsgifanim.cpp,v $
 * Revision 1.1  1999/10/07 23:05:24  daniel
 * Initial version
 *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "owtcgien.h"
#include "cpl_error.h"
#include "cpl_string.h"

/*---------------------------------------------------------------------
 * Prototype for whirlgif_main().
 * WHIRLGIF_MAIN should have been previously defined in the makefile.
 *--------------------------------------------------------------------*/
extern "C" int WHIRLGIF_MAIN(int argc, char *argv[]);

#ifdef _WIN32
#   define TMP_ENV   "\\temp\\cgienv.env"
#else
#   define TMP_ENV   "/tmp/cgienv.env"
#endif

void    PrintInfo();


extern "C" void MyErrorHandler(CPLErr nErrClass, int nErrno, const char *msg)
{
    fprintf(stdout, "Content-type: text/plain\n\n");
    fprintf(stdout, "ERROR %d: %s\n", nErrno, msg);

    exit(1);
}


/**********************************************************************
 *                       main()
 *
 * Return an animated GIF built from a set of SPANS Web Server GIF slides.
 *
 * The parameters are passed through the query string part of the URL:
 *
 *  /cgi-bin/wsgifanim?...parameters...
 *
 * The following parameters are accepted:
 * (Corresponding to the command-line whirlgif parameters)
 *
 *      loop=[count]	add the Netscape 'loop' extension.
 *      time=delay	inter-frame timing (can be repeated).
 *      disp=[method]	Set disposal method, see whirlgif.doc
 *      trans=index	set the colormap index 'index' to be transparent
 *      id=file_id      SPANS Web Server cache id of one GIF slide to include
 *                      this param. should be repeated for each slide.
 *
 * See also: whirlgif.doc
 *
 * Note that you can specify several delay statements between files to
 * change the delay between images in the middle of an animation.
 *
 ********************************************************************** */

int main(int argc, char *argv[]) 
{
    char **papszArgv = NULL, **papszParams;
    int i, numParams, nStatus = 0;
    int numFiles = 0;
    char *pszName, *pszValue;

    /*-----------------------------------------------------------------
     * Create a OWTCGIEnv object that will be used to handle CGI stuff.
     * When the program exits the object will be destroyed and any 
     * resources it uses will be freed automagically.
     *----------------------------------------------------------------*/
    OWTCGIEnv oCGIEnv;

    /*-----------------------------------------------------------------
     * Save/Restore saved CGI Env. State for debugging
     *----------------------------------------------------------------*/
    // oCGIEnv.DumpToFile(TMP_ENV);
#ifdef DEBUG
    if (strlen(oCGIEnv.GetEnv("SCRIPT_NAME")) == 0 &&
        strlen(oCGIEnv.GetEnv("PATH_INFO")) == 0 &&
        strlen(oCGIEnv.GetEnv("QUERY_STRING")) == 0 )
        oCGIEnv.InitFromFile(TMP_ENV);
#endif

    CPLSetErrorHandler(MyErrorHandler);

    /* ----------------------------------------------------------
     * Build the argv[] to pass to WHIRLGIF_MAIN(), starting with
     * the current program name.
     * ---------------------------------------------------------- */
    if (argc > 0)
        papszArgv = CSLAddString(NULL, argv[0]);
    else
        papszArgv = CSLAddString(NULL, "wsgifanim");

    /* ----------------------------------------------------------
     * Loop through the QUERY_STRING params and fill the argv[]
     * ---------------------------------------------------------- */
    papszParams = oCGIEnv.GetAllFormParams();
    numParams = CSLCount(papszParams);

    for(i=0; i<numParams; i++)
    {
        /* ------------------------------------------------------
         * Extract the name and value part for this param.
         * (It's OK to modify the source strings since we won't reuse them.)
         * ------------------------------------------------------ */
        pszName = pszValue = papszParams[i];
        while(*pszValue != '\0' && *pszValue != '=')
            pszValue++;
        if (*pszValue == '=')
        {
            *pszValue = '\0';
            pszValue++;
        }

        /* ------------------------------------------------------
         * ... and convert it into something whirlgif will understand
         * ------------------------------------------------------ */
        if (EQUAL(pszName, "LOOP"))
        {
            papszArgv = CSLAddString(papszArgv, "-loop");
            papszArgv = CSLAddString(papszArgv, pszValue);
        }
        else if (EQUAL(pszName, "TIME"))
        {
            papszArgv = CSLAddString(papszArgv, "-time");
            papszArgv = CSLAddString(papszArgv, pszValue);
        }
        else if (EQUAL(pszName, "DISP"))
        {
            papszArgv = CSLAddString(papszArgv, "-time");
            papszArgv = CSLAddString(papszArgv, pszValue);
        }
        else if (EQUAL(pszName, "TRANS"))
        {
            papszArgv = CSLAddString(papszArgv, "-time");
            papszArgv = CSLAddString(papszArgv, pszValue);
        }
        else if (EQUAL(pszName, "ID"))
        {
            numFiles++;
            papszArgv = CSLAddString(papszArgv, pszValue);
        }
    }

    if (numFiles == 0)
    {
        /* ------------------------------------------------------
         * In case argv[] is empty, then just return program info.
         * ------------------------------------------------------ */
        PrintInfo();
    }
    else
    {
        /* ------------------------------------------------------
         * Pass control to Whirlgif to do the actual work
         * ------------------------------------------------------ */
        printf("Content-Type: image/gif\n\n");
        nStatus = WHIRLGIF_MAIN(CSLCount(papszArgv), papszArgv);
    }
    CSLDestroy(papszArgv);

    return nStatus;
}


/**********************************************************************
 *                       PrintInfo()
 *
 * Return Version and Copyright information to stdout.
 ********************************************************************** */

void PrintInfo()
{
    printf("Content-Type: text/plain\n\n");

    printf("GIF Slides Animator (for SPANS Web Server slides)\n\n");
    printf("Copyright (c) 1999, Daniel Morissette (danmo@videotron.ca)\n\n");
    printf("Uses Whirlgif by Kevin Kadow (kadokev@msg.net) \n");
    printf("(http://www.msg.net/utility/whirlgif/)\n");
    printf("\n");
    printf("$Revision: 1.1 $ (Built %s %s)\n", __DATE__, __TIME__);
}

