/**********************************************************************
 * $Id: dbftable.cpp,v 1.6 2001/09/14 20:08:00 assefa Exp $
 *
 * Name:     dbftable.cpp
 * Project:  OWT - DBF File access
 * Purpose:  Implementation of the DBFTable class
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: dbftable.cpp,v $
 * Revision 1.6  2001/09/14 20:08:00  assefa
 * Remove front/trailing white spaces in field names.
 *
 * Revision 1.5  2000/06/15 20:06:55  daniel
 * Ported to Windows
 *
 * Revision 1.4  1999/05/28 21:25:05  daniel
 * Little changes to port to VC++
 *
 * Revision 1.3  1999/05/28 16:53:45  daniel
 * Trim spaces when reading field values
 *
 * Revision 1.2  1998/12/09 02:08:02  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:20  daniel
 * Initial revision
 *
 * Revision 1.2  1998/10/30 16:40:06  daniel
 * Added RCS header
 *
 **********************************************************************/

#include "cpl_error.h"
#include "dbftable.h"

#include <malloc.h>
#include <math.h>
#include <string.h>
#include <fcntl.h>  /* O_RDONLY     */

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#else
#include <unistd.h> /* close()      */
#endif

/*=====================================================================
                            DBFTable class
 =====================================================================*/

/**********************************************************************
 *                       DBFTable::DBFTable()
 *
 **********************************************************************/

DBFTable::DBFTable()
{
    m_dbh = NULL;
    m_dbfd = -1;
    m_nCurRecordId = -1;
    m_pszCurRecord = NULL;
    m_pszCurFieldBuffer = NULL;
}

/**********************************************************************
 *                       DBFTable::~DBFTable
 *
 ********************************************************************** */

DBFTable::~DBFTable()
{
    if (m_dbh != NULL)
        Close();
}

/**********************************************************************
 *                       DBFTable::Open()
 *
 * Try to open a DBF file.  
 * Returns TRUE on success of FALSE if it fails.
 *
 * For now, DBF files are always opened read-only.
 **********************************************************************/

GBool DBFTable::Open(const char *pszFname)
{
    /* Check if file is already opened */
    if (m_dbh != NULL)
        return FALSE;
    
    /* Open the physical file... on Windows, we have to force binary mode */
#ifdef _WIN32
    m_dbfd = open(pszFname, O_RDONLY | O_BINARY);
#else
    m_dbfd = open(pszFname, O_RDONLY);
#endif
    if (m_dbfd < 0)
        return FALSE;

    /* Fetch DBF header */
    m_dbh = get_dbf_head(m_dbfd);
    if (m_dbh == NULL)
    {
        close(m_dbfd);
        m_dbfd = -1;
        return FALSE;
    }

    /* Allocate static buffer for GetFieldValue() */
    m_pszCurFieldBuffer = (char *)calloc(m_dbh->db_rlen+1, sizeof(char));
    if (m_pszCurFieldBuffer == NULL)
    {
        Close();
        return FALSE;
    }

    return TRUE;
}

/**********************************************************************
 *                       DBFTable::Close()
 *
 * Close opened DBF table.
 **********************************************************************/

void DBFTable::Close()
{
    if (m_dbh == NULL)
        return;

    /* Close file */
    free_dbf_head(m_dbh);
    m_dbh = NULL;

    close(m_dbfd);
    m_dbfd = -1;

    /* Reset member variables */
    m_nCurRecordId = -1;

    if (m_pszCurRecord)
    {
        free(m_pszCurRecord);
        m_pszCurRecord = NULL;
    }

    if (m_pszCurFieldBuffer)
    {
        free(m_pszCurFieldBuffer);
        m_pszCurFieldBuffer = NULL;
    }

}


/**********************************************************************
 *                       DBFTable::GetNumFields()
 *
 * Return number of fields in one record or -1 if table not opened.
 **********************************************************************/

int  DBFTable::GetNumFields()
{
    if (m_dbh == NULL)
        return -1;

    return m_dbh->db_nfields;
}

/**********************************************************************
 *                       DBFTable::GetNumRecords()
 *
 * Return number of Records in table or -1 if table not opened.
 **********************************************************************/

int  DBFTable::GetNumRecords()
{
    if (m_dbh == NULL)
        return -1;

    return m_dbh->db_records;
}


char* RemoveWhiteSpace( char *pszString )
{
    char* pszReadPos = pszString;
    char* pszWritePos = pszString;

 
    while (*pszReadPos)
    {
        if ((*pszReadPos) == ' ')
        {
            pszReadPos++;
        }
        else
        {
            *pszWritePos = *pszReadPos;
            pszWritePos++;
            pszReadPos++;
        }
    }

    *pszWritePos = '\0';

    return pszString;
}

/**********************************************************************
 *                       DBFTable::GetFieldIndex()
 *
 * Look for a field by its name and return its index or -1 if not found.
 *
 * Note that field indices are 1 based.
 **********************************************************************/

int  DBFTable::GetFieldIndex(const char *pszFieldName)
{
    int     nIndex;
    GBool    bFound = FALSE;

    if (m_dbh == NULL || m_dbh->db_nfields < 1)
        return -1;

    for (nIndex=1; nIndex<=m_dbh->db_nfields; nIndex++)
    {
        if (EQUAL(pszFieldName,RemoveWhiteSpace((char*)m_dbh->db_fields[nIndex-1].db_fname)))
        {
            bFound = TRUE;
            break;
        }
    }

    if (!bFound)
        nIndex = -1;

    return nIndex;
}


/**********************************************************************
 *                       DBFTable::GetFieldName()
 *
 * Return the specified field's name (static buffer ptr) or NULL if
 * field not found.
 *
 * Note that field indices are 1 based.
 **********************************************************************/

const char *DBFTable::GetFieldName(int nFieldIndex)
{
    if (m_dbh == NULL || nFieldIndex < 1 || nFieldIndex > m_dbh->db_nfields)
        return NULL;

    return (char*)m_dbh->db_fields[nFieldIndex-1].db_fname;
}
        
/**********************************************************************
 *                       DBFTable::GetFieldType()
 *
 * Return the xbase field type: 
 *  C = Character
 *  D = Date
 *  N = Numeric
 *  F = Float
 *  L = Logical
 *  M = Memo (.DBT block number)
 *  B = Binary (.DBT block number)
 *  G = General (.DBT block number)
 *
 * Returns '?' if the field type could not be established.
 **********************************************************************/

char DBFTable::GetFieldType(const char *pszFieldName)
{
    return GetFieldType(GetFieldIndex(pszFieldName));
}

char DBFTable::GetFieldType(int nFieldIndex)
{
    if (m_dbh == NULL || nFieldIndex < 1 || nFieldIndex > m_dbh->db_nfields)
        return '?';

    return m_dbh->db_fields[nFieldIndex-1].db_type;
}


/**********************************************************************
 *                       DBFTable::GotoRecord()
 *
 * (Private method)
 *
 * Move the cursor to the specified record number.  Does nothing if
 * already pointing to the right record.  The record information will
 * be stored in m_pszCurRecord.
 *
 * Note that record indices are 1 based.
 *
 * Returns TRUE on success or FALSE otherwise.
 **********************************************************************/

GBool DBFTable::GotoRecord(int nRecordId)
{
    if (m_dbh == NULL || nRecordId < 1 || nRecordId > m_dbh->db_records)
        return FALSE;

    if (m_nCurRecordId == nRecordId)
        return TRUE;    /* Already pointing to the right record */

    /* Read the record data (string) and store it in the m_pszCurRecord
     * buffer for later use.
     */
    if (m_pszCurRecord)
        free(m_pszCurRecord);
    m_pszCurRecord = (char *)get_dbf_record(m_dbh, nRecordId);
    m_nCurRecordId = nRecordId;

    if (m_pszCurRecord == NULL)
    {
        /* For some reason, the record could not be read... */
        m_nCurRecordId = -1;
        return FALSE;
    }

    return TRUE;
}

/**********************************************************************
 *                       DBFTable::GetFieldValue()
 *
 * Read the specified value.  
 * Return a pointer to a static buffer or NULL if request failed.
 *
 * Note that Record and Field indices are 1 based.
 *
 * The static buffer's contents is guaranteed to be valid 
 * only until the next call to GetFieldValue().
 **********************************************************************/

const char *DBFTable::GetFieldValue(int nRecordId, const char *pszFieldName)
{
    return GetFieldValue(nRecordId, GetFieldIndex(pszFieldName));
}

const char *DBFTable::GetFieldValue(int nRecordId, int nFieldIndex)
{
    dbfield_t   *cur_f;
    int nLen;
    const char *pszRetValue;

    if (m_dbh == NULL || nFieldIndex < 1 || nFieldIndex > m_dbh->db_nfields)
        return NULL;

    /* move table pointer to specified record */
    if (GotoRecord(nRecordId) == FALSE)
        return NULL;

    /* Extract the field value from the local buffer m_pszCurRecord */
    cur_f = &(m_dbh->db_fields[nFieldIndex-1]);
    CPLAssert(cur_f);
    CPLAssert(m_pszCurFieldBuffer);

    strncpy(m_pszCurFieldBuffer, 
            m_pszCurRecord+cur_f->db_foffset, cur_f->db_flen);
    m_pszCurFieldBuffer[cur_f->db_flen] = '\0';

    /* Remove trailing spaces. */
    for( nLen = cur_f->db_flen-1 ; 
         nLen >= 0 && m_pszCurFieldBuffer[nLen] == ' ';
         nLen--)
    {
        m_pszCurFieldBuffer[nLen] = '\0';
    }

    pszRetValue = m_pszCurFieldBuffer;

    /* For a numeric value, skip spaces at the beginning of the value as well 
     */
    if (m_dbh->db_fields[nFieldIndex-1].db_type == 'N')
    {
        while(*pszRetValue == ' ')
            pszRetValue++;
    }

    /* Return a pointer to a static buffer... valid only until next call. */
    return pszRetValue;
}



#ifdef DBFTABLE_SELF_TEST
/*=====================================================================
                         DBFTable test routines
 =====================================================================*/

/**********************************************************************
 *                       _DBFTableInfo()
 *
 * Print misc. table info.
 *
 * Returns TRUE on success or FALSE if any error happened.
 **********************************************************************/
static GBool _DBFTableInfo(DBFTable &oTable)
{
    int i;

    printf("Table info:\n\n");
    printf("  Num Fields:  %d\n", oTable.GetNumFields());
    printf("  Num Records: %d\n", oTable.GetNumRecords());

    printf("Field names and types:\n");

    for (i=1; i<=oTable.GetNumFields(); i++)
    {
        printf("  %2.2d: (%c) %s\n", i, oTable.GetFieldType(i), 
                                        oTable.GetFieldName(i));
    }

    printf("\n\n");

    return TRUE;
}

/**********************************************************************
 *                       _DBFPrintRecord()
 *
 * Print values for specified record.
 *
 * Returns TRUE on success or FALSE if any error happened.
 **********************************************************************/
static GBool _DBFPrintRecord(DBFTable &oTable, int nRecordId)
{
    GBool        bSuccess = TRUE;
    const char *pszFieldValue, *pszFieldName;
    int         i;

    printf("Record # %d:\n", nRecordId);

    for (i=1; bSuccess && i<=oTable.GetNumFields(); i++)
    {
        pszFieldName = oTable.GetFieldName(i);
        pszFieldValue = oTable.GetFieldValue(nRecordId, pszFieldName);
        if (pszFieldName == NULL || pszFieldValue == NULL)
        {
            bSuccess = FALSE;
            printf("Failed reading field value for record %d, field %d\n", 
                                                        nRecordId, i);
        }
        else
            printf("  [%s] = [%s]\n", pszFieldName, pszFieldValue);
    }

    printf("\n");

    return bSuccess;
}

/**********************************************************************
 *                       testDBFTable()
 *
 * Exercise and test the DBFTable class.
 **********************************************************************/
GBool testDBFTable(const char *pszFname)
{
    GBool bSuccess = TRUE;
    DBFTable    oTable;

    bSuccess = oTable.Open(pszFname);

    if (bSuccess)
        bSuccess = _DBFTableInfo(oTable);

    if (bSuccess)
        bSuccess = _DBFPrintRecord(oTable, 1);

    if (bSuccess)
        bSuccess = _DBFPrintRecord(oTable, oTable.GetNumRecords());

    if (bSuccess)
        bSuccess = _DBFPrintRecord(oTable, oTable.GetNumRecords()/2);

    if (bSuccess)
        bSuccess = _DBFPrintRecord(oTable, oTable.GetNumRecords()/3);

    if (bSuccess)
        printf("DBFTable tests succeded.\n\n");
    else
        printf("DBFTable tests failed\n\n");

    oTable.Close();

    return bSuccess;
}


#endif /* DBFTABLE_SELF_TEST */

