/**********************************************************************
 * $Id: htblock.cpp,v 1.8 2000/06/15 20:06:55 daniel Exp $
 *
 * Name:     htbloc.cpp
 * Project:  OWT HTML manipulation library
 * Purpose:  Implementation of the OWTHtBlock class.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: htblock.cpp,v $
 * Revision 1.8  2000/06/15 20:06:55  daniel
 * Ported to Windows
 *
 * Revision 1.7  1999/11/19 16:37:38  stephane
 * comment out default parameter in OWTExpandEnvVars
 *
 * Revision 1.6  1999/07/14 04:14:56  daniel
 * Added special case in SetChild() for OWTBLOCK_UNIQUE_CHILD_NAME,
 * + made OWTExpandEnvVars() available as a function
 *
 * Revision 1.3  1999/05/28 16:54:55  daniel
 * Use CPL, added SetChildText() and ExpandEnvVars()
 *
 * Revision 1.2  1998/12/09 02:08:03  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:16  daniel
 * Initial revision
 *
 *
 **********************************************************************/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "owtht.h"

/*=====================================================================
                        OWTHtBlock class.
 =====================================================================*/

/**********************************************************************
 *                       OWTHtBlock::OWTHtBlock()
 *
 **********************************************************************/
OWTHtBlock::OWTHtBlock(const char *pszName /*=NULL*/, 
                     const char *pszType /*=NULL*/, 
                     const char *pszOptions /*=NULL*/)
{
    m_pszName = m_pszType = m_pszOptions = NULL;
    m_papszText = NULL;
    m_bParsed = FALSE;
    m_numChildren = 0;
    m_papoChildren = NULL;

    if (pszName)
        m_pszName = CPLStrdup(pszName);

    if (pszType)
        m_pszType = CPLStrdup(pszType);

    if (pszOptions)
        m_pszOptions = CPLStrdup(pszOptions);
}

OWTHtBlock::OWTHtBlock(OWTTagInfo *psTagInfo)
{
    if (psTagInfo)
        OWTHtBlock(psTagInfo->pszName, psTagInfo->pszType, 
                  psTagInfo->pszOptions);
    else
        OWTHtBlock();
}

/**********************************************************************
 *                       OWTHtBlock::~OWTHtBlock()
 *
 **********************************************************************/
OWTHtBlock::~OWTHtBlock()
{
    if (m_pszName)
        CPLFree(m_pszName);
    if (m_pszType)
        CPLFree(m_pszType);
    if (m_pszOptions)
        CPLFree(m_pszOptions);

    if (m_papoChildren)
    {
        for (int i=0; i<m_numChildren; i++)
            delete m_papoChildren[i];
        CPLFree(m_papoChildren);
    }

    CSLDestroy(m_papszText);
}

/**********************************************************************
 *                       OWTHtBlock::SetText()
 *
 * Set the text contents of the current object.  This will replace
 * any previous contents, and reset the object to a not-parsed state.
 *
 * SetText() defaults to allocate its own copy of papszSrc, but if
 * bMakeCopy is passed as FALSE, then the current object becomes the
 * owner of the StringList... and it will take care of freeing it when
 * time comes.
 **********************************************************************/
void OWTHtBlock::SetText(char **papszSrc, GBool bMakeCopy /* =TRUE */)
{
    if (m_papszText)
        CSLDestroy(m_papszText);

    if (bMakeCopy)
        m_papszText = CSLDuplicate(papszSrc);
    else
        m_papszText = papszSrc;

    if (m_papoChildren)
    {
        for (int i=0; i<m_numChildren; i++)
            delete m_papoChildren[i];
        CPLFree(m_papoChildren);
        m_papoChildren = NULL;
    }
    m_numChildren = 0;
    m_bParsed = FALSE;
}

/**********************************************************************
 *                       OWTHtBlock::SetText()
 *
 * Same as above, but takes a single string as input.
 **********************************************************************/

void OWTHtBlock::SetText(char *pszNewText)
{
    char *apszText[2];

    /* Create a temporary StringList and call the other SetText()
     */
    apszText[0] = pszNewText;
    apszText[1] = NULL;

    SetText(apszText);
}


/**********************************************************************
 *                       OWTHtBlock::SetChildText()
 *
 * Set the text contents of a child of this object.  If the named
 * child does not exist, then the function simply does nothing.
 *
 * SetChildText() defaults to allocate its own copy of papszSrc, but if
 * bMakeCopy is passed as FALSE, then the current object becomes the
 * owner of the StringList... and it will take care of freeing it when
 * time comes.
 **********************************************************************/
void OWTHtBlock::SetChildText(const char *pszChildName,
                             char **papszSrc, GBool bMakeCopy /* =TRUE */)
{
    OWTHtBlock   *pChild;

    pChild = GetChild(pszChildName);
    if (pChild)
    {
        pChild->SetText(papszSrc, bMakeCopy);
    }
    else if (bMakeCopy == FALSE)
    {
        /* Child does not exist, but we must free the stringlist since
         * the caller expects us to take care of it!
         */
        CSLDestroy(papszSrc);
    }
}


void OWTHtBlock::SetChildText(const char *pszChildName,
                              const char *pszText)
{
    char **papszText = NULL;

    /* Alloc a new string list, store the line of text in it, and 
     * pass it to the other SetChildText(), asking it to take control
     * of the stringlist (to free it once it's done with it)
     */
    papszText = CSLAddString(papszText, pszText);
    SetChildText(pszChildName, papszText, FALSE );
}

/**********************************************************************
 *                       OWTHtBlock::Draw()
 *
 * Print the contents of the block to the specified file.
 *
 * Default output file is stdout.
 **********************************************************************/
void OWTHtBlock::Draw(FILE *fpOut /*=NULL*/)
{
    char **ppszLine;
    int  nLenTagStart, iChild;

    if (m_papszText == NULL)
        return; /* Nothing to do! */

    if (fpOut == NULL)
        fpOut = stdout;

    if (!m_bParsed)
    {
        /* Block has not been parsed yet... just return its text contents
         */
        CSLPrint(m_papszText, fpOut);
    }
    else
    {
        nLenTagStart = strlen(OWTTAG_START);
        /* Block has been parsed.  We have to rebuild the output
         */
        ppszLine = m_papszText;
        iChild = 0;
        while(*ppszLine != NULL)
        {
            if (EQUALN(*ppszLine, OWTTAG_START, nLenTagStart))
            {
                /* Found a block tag... expand the corresponding block
                 */
                if (EQUAL(m_papoChildren[iChild]->GetName(), 
                          OWTBLOCK_UNIQUE_CHILD_NAME))
                {
                    /* Block is called OWTBLOCK_UNIQUE_CHILD_NAME... we 
                     * will print the block only, without the start/end tags.
                     */
                    m_papoChildren[iChild++]->Draw(fpOut);
                    /* Move pointer to end of text array to force end of loop.
                     */
                    while(*ppszLine != NULL)
                        ppszLine++;
                }
                else
                {
                    fprintf(fpOut, "%s\n", *ppszLine++);
                    m_papoChildren[iChild++]->Draw(fpOut);
                    fprintf(fpOut, "%s\n", *ppszLine++);
                }
            }
            else
            {
                /* Just a simple line... print it directly.
                 */
                fprintf(fpOut, "%s\n", *ppszLine++);
            }
        }
    }/*bParsed==TRUE*/
}

/**********************************************************************
 *                       OWTHtBlock::Parse()
 *
 * Parse the dumb text version of this HtBlock into a smart version
 * that we can edit/modify, etc.
 **********************************************************************/
void OWTHtBlock::Parse()
{
    OWTTagInfo *psTagInfo;
    int     iLine;
    char    **papszChildText;

    if (m_bParsed)
        return; /* Already parsed... nothing to do! */

    /* Make sure tags are organized in a way that they can be
     * easily parsed.
     */
    m_papszText = OWTPreprocessTags(m_papszText);

    m_numChildren = 0;
    iLine = 0;
    while((psTagInfo = OWTGetNextTagInfo(m_papszText, iLine, -1)) != NULL)
    {
        /* Found a new child... add it to the list and extract its contents
         * from the m_papszText
         */
        m_numChildren++;

        if (m_papoChildren)
        {
            m_papoChildren = (OWTHtBlock **)CPLRealloc(m_papoChildren,
                                            m_numChildren*sizeof(OWTHtBlock*));
        }
        else
        {
            m_papoChildren = (OWTHtBlock **)CPLMalloc(m_numChildren*
                                                    sizeof(OWTHtBlock*));
        }

        if (m_papoChildren == NULL)
        {
            /* Error: Out of memory!!! */
            CPLError(CE_Fatal, CPLE_OutOfMemory,
                     "Parse(): Unable to allocate memory.");
            return;
        }

        /* Create new child
         */
        m_papoChildren[m_numChildren-1] = new OWTHtBlock(psTagInfo->pszName,
                                                       psTagInfo->pszType,
                                                       psTagInfo->pszOptions);
        /* Extract the lines of text corresponding to the current child from
         * m_papszText and pass them to the child.
         */
        papszChildText = NULL;
        m_papszText = CSLRemoveStrings(m_papszText, 
                                            psTagInfo->nFirstLine+1,
                                            psTagInfo->nNumLines-2,
                                            &papszChildText);
        m_papoChildren[m_numChildren-1]->SetText(papszChildText, FALSE);
            
        /* Continue scanning the source text with the line that follows
         */
        iLine = psTagInfo->nFirstLine + 2;

        OWTFreeTagInfo(psTagInfo);
    }

    m_bParsed = TRUE;
}

/**********************************************************************
 *                       OWTHtBlock::GetChild()
 *
 * Looks for a child with the specified name in the current HtBlock.
 * If it is found then return a reference to it (a OWTHtBlock object)
 * or return NULL if not found.
 **********************************************************************/
OWTHtBlock *OWTHtBlock::GetChild(const char *pszChildName)
{
    OWTHtBlock  *poChild = NULL;
    int         iChild;

    /* First, make sure the current block has been parsed
     */
    Parse();

    /* Scan the list of children...
     */
    for(iChild=0; iChild < m_numChildren; iChild++)
    {
        if (EQUAL(m_papoChildren[iChild]->GetName(), pszChildName))
        {
            /* Found it! */
            poChild = m_papoChildren[iChild];
            break;
        }
    }

    return poChild;
}

/**********************************************************************
 *                       OWTHtBlock::SetChild()
 *
 * Looks for a child with the specified name in the current HtBlock
 * and replace the current object with the new one.  
 * Note that the object that is passed will become property of this
 * HtBlock object, so the caller should not modify or delete it.
 *
 * The OWTHtBlock child named OWTBLOCK_UNIQUE_CHILD_NAME has a 
 * special behavior: use this name to pass a block that will be
 * encapsulated by the current block.  Doing a SetChild()
 * with that name will completely replace the contents of the block.
 *
 * If it is found then return a reference to it (a OWTHtBlock object)
 * or return NULL if not found.
 **********************************************************************/
OWTHtBlock *OWTHtBlock::SetChild(const char *pszChildName,
                                 OWTHtBlock *poNewChild)
{
    int         iChild;

    if (EQUAL(pszChildName, OWTBLOCK_UNIQUE_CHILD_NAME))
    {
    /*-----------------------------------------------------------------
     * If child is called OWTBLOCK_UNIQUE_CHILD_NAME then it replaces the
     * whole block contents...
     *----------------------------------------------------------------*/
        char **papszText = NULL;

        /* Erase the whole block... keep only the necessary lines to trigger
         * redrawing the unique child.
         */
        papszText = CSLAppendPrintf(papszText, OWTTAG_START);
        papszText = CSLAppendPrintf(papszText, OWTTAG_END);

        SetText(papszText, FALSE);

        m_numChildren = 1;
        m_papoChildren = (OWTHtBlock **)CPLMalloc(m_numChildren*
                                                    sizeof(OWTHtBlock*));
        m_papoChildren[0] = poNewChild;
        poNewChild->SetName(pszChildName);

        /* Mark this block as parsed */
        m_bParsed = TRUE;

        return poNewChild;
    }
    else
    {
    /*-----------------------------------------------------------------
     * It's a normal child... find and replace it.
     *----------------------------------------------------------------*/

        /* First, make sure the current block has been parsed
         */
        Parse();

        /* Scan the list of children...
         */
        for(iChild=0; iChild < m_numChildren; iChild++)
        {
            if (EQUAL(m_papoChildren[iChild]->GetName(), pszChildName))
            {
                /* Found it! */
                delete m_papoChildren[iChild];
                m_papoChildren[iChild] = poNewChild;
                poNewChild->SetName(pszChildName);
                return poNewChild;
            }
        }
    }

    return NULL;
}

/**********************************************************************
 *                       OWTHtBlock::GetName()
 *
 * Return this block's name, or "" if no name has been set for it.
 * The returned value is should not be modified nor freed by the caller.
 **********************************************************************/
const char *OWTHtBlock::GetName()
{
    if (m_pszName)
        return m_pszName;
    else
        return "";
}


/**********************************************************************
 *                       OWTHtBlock::SetName()
 *
 * Set a new value for this object's name.
 **********************************************************************/
void OWTHtBlock::SetName(const char *pszNewName)
{
    CPLFree(m_pszName);
    m_pszName = CPLStrdup(pszNewName);
}


/**********************************************************************
 *                       OWTHtBlock::ExpandEnvVars()
 *
 * Look for all occurences of env. vars ("$(VAR_NAME)") in the curent
 * document, and replace them with the value returned by gettenv()
 * for each variable.
 *
 * If papszValues != NULL, then the values in this stringlist of 
 * name=value pairs will have priority over the system's env. vars.
 **********************************************************************/
void    OWTHtBlock::ExpandEnvVars(OWTCGIEnv *poCGIEnv /*=NULL*/,
                                  char **papszValues /*=NULL*/)
{
    OWTExpandEnvVars(m_papszText, poCGIEnv, papszValues);
}


/**********************************************************************
 *                       OWTExpandEnvVars()
 *
 * Look for all occurences of env. vars ("$(VAR_NAME)") in the 
 * stringlist, and replace them with the value returned by gettenv()
 * for each variable.
 *
 * This function modifies the lines in the original stringlist directly.
 *
 * If papszValues != NULL, then the values in this stringlist of 
 * name=value pairs will have priority over the system's env. vars.
 **********************************************************************/
#define MAX_VAR_LEN  100
void OWTExpandEnvVars(char **papszSrcText,
                      OWTCGIEnv *poCGIEnv /*=NULL*/,
                      char **papszValues /*=NULL*/)
{
    char szVarName[MAX_VAR_LEN+1], *pszCurPtr, *pszVarFound, *pszLine;
    int  iLine, nVarLen;
    GBool bFoundVar;
    const char *pszVarValue;

    if (papszSrcText == NULL)
        return;

    for(iLine=0; papszSrcText[iLine]; iLine++)
    {
        pszLine = pszCurPtr = papszSrcText[iLine];

        while((pszVarFound = strstr(pszCurPtr, "$(")) != NULL)
        {
            pszCurPtr = pszVarFound + 2;
            bFoundVar = FALSE;

            /*--------------------------------------------------------
             * Read variable name
             *-------------------------------------------------------*/
            for(nVarLen=0; 
                nVarLen < MAX_VAR_LEN && *pszCurPtr != '\0' ; nVarLen++)
            {
                if (*pszCurPtr == ')')
                {
                    /* Finished reading var. name ... */
                    szVarName[nVarLen] = '\0';
                    pszCurPtr++;
                    bFoundVar = TRUE;
                    break;
                }
                else if (*pszCurPtr == '_' || isalnum( *pszCurPtr ) )
                {
                    /* Still reading var. name ... continue the loop */
                    szVarName[nVarLen] = *pszCurPtr;
                    pszCurPtr++;
                    continue;
                }
                else
                {
                    /* Invalid variable ... abort reading var. name... we
                     * will continue scanning the line from here
                     */
                    szVarName[0] = '\0';
                    break;
                }
            }
            szVarName[nVarLen] = '\0';

            /*--------------------------------------------------------
             * Check for variable value.
             * First check in papszValues, then in the OWTCGIEnv, and
             * finally, if stil not found then call getenv() directly
             *-------------------------------------------------------*/
            pszVarValue = NULL;
            if (bFoundVar && nVarLen > 0)
            {
                if (papszValues)
                    pszVarValue = CSLFetchNameValue(papszValues, szVarName);

                if ( pszVarValue == NULL )
                {
                    if (poCGIEnv)
                        pszVarValue = poCGIEnv->GetEnv(szVarName);
                    else
                        pszVarValue = getenv(szVarName);
                }
            }
            
            /*--------------------------------------------------------
             * If we found something then replace with the value
             *-------------------------------------------------------*/
            if (pszVarValue)
            {
                int nNewLen;
                nNewLen = (pszVarFound-pszLine) + 
                                     strlen(pszVarValue) + strlen(pszCurPtr);

                papszSrcText[iLine] = (char*)CPLMalloc((nNewLen+1)*
                                                            sizeof(char));
                *pszVarFound = '\0';

                strcpy(papszSrcText[iLine], pszLine);
                strcat(papszSrcText[iLine], pszVarValue);
                strcat(papszSrcText[iLine], pszCurPtr);

                /* Set vars to continue scanning this line at the next 
                 * pass in the loop 
                 */
                pszCurPtr = papszSrcText[iLine] + (pszVarFound-pszLine) + 
                                                         strlen(pszVarValue);
                CPLFree(pszLine);
                pszLine = papszSrcText[iLine];
            }

        }/*while ... scanning one line*/

    }/* for iLine */

}

