/**********************************************************************
 * $Id: owtgdchart.cpp,v 1.3 2000/11/23 22:38:50 daniel Exp $
 *
 * Name:     owtgdchart.cpp
 * Project:  OWT - Chart building block
 * Purpose:  Implementation of the OWTGDChart class.
 * Note:     This class uses the GDChart and GifDraw (gd1.3) library.
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998-2000, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtgdchart.cpp,v $
 * Revision 1.3  2000/11/23 22:38:50  daniel
 * In gdchart, gifwidth/gifheight changed to imgwidth/imgheight
 *
 * Revision 1.2  1999/05/28 21:35:31  daniel
 * *** empty log message ***
 *
 * Revision 1.1  1999/05/28 21:33:49  daniel
 * Initial revision
 *
 **********************************************************************/

#include "cpl_conv.h"
#include "cpl_string.h"

#include "owtgdchart.h"

#include "gdfontt.h"
#include "gdfonts.h"

#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#ifdef WIN32
#include <io.h>     /* _setmode()   */
#include <fcntl.h>  /* O_BINARY     */
#endif

/*=====================================================================
                            OWTGDChart class
 =====================================================================*/

/**********************************************************************
 *                       OWTGDChart::OWTGDChart()
 *
 **********************************************************************/

OWTGDChart::OWTGDChart(GDC_CHART_T nChartType)
{
    m_pfValues = NULL;
    m_numLines = 0;
    m_numValPerLine = 0;

    m_numYAxisDiv = 5;
    m_bHorGrid = FALSE;
    m_bXAxisTickMarks = TRUE;
    m_bYAxisTickMarks = TRUE;

    m_papszXValueLabels = NULL;

    /*-----------------------------------------------------------------
     * Alloc a GDCChart struct with default size of 400x300
     *----------------------------------------------------------------*/
    psGDC = GDCChartAlloc(nChartType, 400, 300);
}

/**********************************************************************
 *                       OWTGDChart::~OWTGDChart()
 *
 ********************************************************************** */

OWTGDChart::~OWTGDChart()
{
    if (m_pfValues != NULL)
        CPLFree(m_pfValues);

    CSLDestroy(m_papszXValueLabels);

    GDCChartDestroy(psGDC);
}

/**********************************************************************
 *                       OWTGDChart::SetSize()
 *
 * Set new chart size.
 ********************************************************************** */
void OWTGDChart::SetSize(int nWidth, int nHeight)
{
    if (nWidth > 20 && nHeight > 20)
    {
        psGDC->imgwidth = nWidth;
        psGDC->imgheight = nHeight;
    }
}

/**********************************************************************
 *                       OWTGDChart::SetValues()
 *
 * Use this function to pass the array of values to the chart.
 *
 * SetValues() makes its own copy of the array of values.  For multiline
 * charts, the values for the first line come first in the array,
 * followed by the values for the second line, ...
 ********************************************************************** */
void OWTGDChart::SetValues(int numLines, int numValPerLine, double *pValues)
{
    int    i, nValues;

    m_numLines = numLines;
    m_numValPerLine = numValPerLine;

    nValues = numLines * numValPerLine;
    if (nValues == 0) 
        return;

    /* Create our own copy of the array of values */
    if (m_pfValues != NULL)
        CPLFree(m_pfValues);
    m_pfValues = (float*)CPLMalloc(nValues*sizeof(float));
    
    for(i=0; i<nValues; i++)
        m_pfValues[i] = (float)pValues[i];

}

/**********************************************************************
 *                       OWTGDChart::SetValues()
 *
 * Same as above, but takes a NULL-terminated array of strings as input.
 ********************************************************************** */
void OWTGDChart::SetValues(int numLines, int numValPerLine, 
                        char **papszValues)
{
    int    i, nValues;
    double *pValues;

    nValues = numLines * numValPerLine;
    if (nValues == 0 || papszValues == NULL) 
        return;

    /* Create an array of doubles and pass it to SetValues() above and
     * SetValues() will make its own copy of the array... I realize this
     * is not very efficient... but it will be simpler this way for now.
     */
    pValues = (double*)CPLCalloc(nValues,sizeof(double));
    for(i=0; i<nValues && papszValues[i] != NULL; i++)
        pValues[i] = atof(papszValues[i]);

    SetValues(numLines, numValPerLine, pValues);

    CPLFree(pValues);
}



/**********************************************************************
 *                       OWTGDChart::Redraw()
 *
 * (Private method)
 *
 * Generate the LINE or BAR chart
 ********************************************************************** */
void OWTGDChart::Redraw(FILE *fpOut)
{

    psGDC->requested_ymin = 0.0;

    GDCChartSetValues(psGDC,
                      m_numValPerLine,
                      m_papszXValueLabels,
                      m_numLines,
                      m_pfValues);

    GDCChartGIF(psGDC, fpOut);

}

/**********************************************************************
 *                       OWTGDChart::RedrawPie()
 *
 * (Private method)
 *
 * Generate the PIE chart
 ********************************************************************** */
void OWTGDChart::RedrawPie(FILE *fpOut)
{
    GDC_RGB_T   *palColors;
    int         *panExpl=NULL;
    float       fOthers;
    int         i;

    /* Set colors for each slice 
     */
    palColors = (GDC_RGB_T*)CPLMalloc(m_numValPerLine*sizeof(GDC_RGB_T));

    for(i=0; i<m_numValPerLine; i++)
    {
        palColors[i] = 0x10000*200*((i+1)&1) + 0x100*200*((i+1)&2)/2 +
                                                  200*((i+1)&4)/4;
    }
    psGDC->pie_Color = palColors;

    /* Assume the last slice is a others category and compute its value
     */
    fOthers = 0.0;
    if (m_numValPerLine > 2)
    {
        fOthers = m_pfValues[m_numValPerLine-1];
        for(i=0; i<m_numValPerLine-1; i++)
            fOthers -= m_pfValues[i];
    }
    if (fOthers > 0.0)
    {
        m_pfValues[m_numValPerLine-1] = fOthers;

        /* Explode the "Others" slice
         */
        panExpl = (int*)CPLCalloc(m_numValPerLine,sizeof(int));
        panExpl[m_numValPerLine-1] = 20;    /* Explode by 20 pixels, other values are 0 */
        psGDC->pie_explode = panExpl;
    }

    /* Set other style parameters
     */
    psGDC->BGColor = 0xFFFFFFL;
    psGDC->pie_EdgeColor = 0x000000L;

    psGDC->_3d_angle = 45;
    //psGDC->_3d_depth = 15;

    psGDC->pie_percent_labels = GDCPIE_PCT_RIGHT;
    psGDC->pie_label_line = TRUE;
    psGDC->pie_label_dist = 15;

    GDCChartSetValues(psGDC,
                      m_numValPerLine,
                      m_papszXValueLabels,
                      m_numLines,
                      m_pfValues);

    GDCChartGIF(psGDC, fpOut);
                      
    CPLFree(palColors);
    CPLFree(panExpl);

    psGDC->pie_Color = NULL;
    psGDC->pie_explode = NULL;
}

/**********************************************************************
 *                       OWTGDChart::SaveSlideGIF()
 *
 ********************************************************************** */
int OWTGDChart::SaveSlideGIF(FILE *fpOut)
{
    /* Output the image (On WIN32 we have to make sure output file is 
     * in binary mode.)
     */
#ifdef WIN32
    if ( _setmode( _fileno( fpOut ), O_BINARY ) == -1)
        CPLError(CE_Failure, CPLE_FileIO, "OWTGDChart::SaveSlideGIF(): "
                      "Cannot set output file mode (%s)",  strerror(errno));
#endif


    /* The call to Redraw*() will generate the chart and send it to specified
     * output file
     */

    psGDC->BGColor =   0xFFFFFFL;      /* BG color   = white */
    psGDC->LineColor = 0x000000L;      /* line color = black */

    psGDC->PlotColor = 0x0000FFL;      /* Bar colors */

    if (psGDC->type == GDC_LINE || psGDC->type == GDC_3DLINE ||
        psGDC->type == GDC_AREA || psGDC->type == GDC_3DAREA ||
        psGDC->type == GDC_BAR  || psGDC->type == GDC_3DBAR)
    {
        Redraw(fpOut);
    }
    else if (psGDC->type == GDC_2DPIE || psGDC->type == GDC_3DPIE)
    {
        RedrawPie(fpOut);
    }
    else
    {
        CPLError(CE_Failure, CPLE_IllegalArg, "Unsupported Chart type!");
    }
    
    fflush(fpOut);

    return 0;
}

/**********************************************************************
 *                       OWTGDChart::SaveSlideGIF()
 *
 * Same as above, but takes the name of a file to create as argument.
 ********************************************************************** */
int OWTGDChart::SaveSlideGIF(const char *pszFname)
{
    FILE *fpOut;
    int nStatus = 0;

    if ((fpOut = fopen(pszFname, "wb")) != NULL)
    {
        nStatus = SaveSlideGIF(fpOut);
        fclose(fpOut);
    }
    else
    {
        CPLError(CE_Failure, CPLE_FileIO, "OWTGDChart::SaveSlideGIF(%s): %s", 
                    pszFname, strerror(errno));
        nStatus = 1;
    }

    return nStatus;
}

/**********************************************************************
 *                       OWTGDChart::SetXValueLabels()
 *
 * Set strings to use for labels along the X axis.
 ********************************************************************** */
void OWTGDChart::SetXValueLabels(char **papszValueLabels)
{
    CSLDestroy(m_papszXValueLabels);
    m_papszXValueLabels = CSLDuplicate(papszValueLabels);
}


/**********************************************************************
 *                       OWTGDChart::GetLegendSamples()
 *
 * Return a MString with the URL for the legend sample for each line.
 ********************************************************************** */
char **OWTGDChart::GetLegendSamples(const char *pszGIFSampleURL,
                                    int nWidth, int nHeight)
{
    int     iLine;
    char    **papszSamples = NULL;

    for(iLine=0; iLine<m_numLines; iLine++)
    {
        papszSamples = CSLAppendPrintf(papszSamples, 
                                "%s/%d/%d/%d/%d/%d",
                                pszGIFSampleURL,
                                nWidth, nHeight, 200*((iLine+1)&1),
                                                 200*((iLine+1)&2)/2,
                                                 200*((iLine+1)&4)/4);
    }
    return papszSamples;
}

