/**********************************************************************
 * $Id: owtht.h,v 1.9 2002/04/15 18:13:34 julien Exp $
 *
 * Name:     owtht.h
 * Project:  OWT HTML manipulation library
 * Purpose:  Header file for the OWTHt* HTML manipulation classes
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1998, 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtht.h,v $
 * Revision 1.9  2002/04/15 18:13:34  julien
 * Add Cell Attributes support and GetCol()/GetRow() fonctions.
 * Replace ',' bf ' ' in TextTable for Quatro (Bug 874)
 *
 * Revision 1.8  2000/06/15 20:06:55  daniel
 * Ported to Windows
 *
 * Revision 1.7  1999/07/15 15:43:41  lcdc
 * Commented out OWTHtBlock::GetText() and GetChildText()... not implem. yet.
 *
 * Revision 1.6  1999/07/14 04:27:23  daniel
 * Modif. OWTTextTable + added OWTExpandEnvVars()
 *
 * Revision 1.5  1999/06/09 19:26:17  daniel
 * Added GetText() and GetChildText() to OWTHtblock
 *
 * Revision 1.4  1999/06/03 02:09:42  daniel
 * Made Get/SetName() virtual
 *
 * Revision 1.3  1999/05/28 17:04:42  daniel
 * Several changes and new classes
 *
 * Revision 1.2  1998/12/09 02:08:05  daniel
 * Converted to use portability library
 *
 * Revision 1.1  1998/12/04 19:15:19  daniel
 * Initial revision
 *
 * Revision 1.2  1998/11/06 03:50:13  daniel
 * First working version of the Charting Demo application
 *
 * Revision 1.1  1998/11/05 04:36:47  daniel
 * Initial revision
 *
 **********************************************************************/

#ifndef _OWTHT_H_INCLUDED_
#define _OWTHT_H_INCLUDED_

#include <stdio.h>

#include "cpl_port.h"
#include "owttags.h"
#include "owtcgien.h"
#include "owtmisc.h"

/*=====================================================================
                       Fuction prototypes
 =====================================================================*/
void OWTExpandEnvVars(char **papszSrcText,
                      OWTCGIEnv *poCGIEnv = NULL,
                      char **papszValues = NULL);

/*---------------------------------------------------------------------
 * The OWTHtBlock child named OWTBLOCK_UNIQUE_CHILD_NAME will have a 
 * special behavior: use this name to pass a block that will be
 * encapsulated by the current block.  Doing a SetChild()
 * with that name will completely replace the contents of the block.
 *--------------------------------------------------------------------*/
#define OWTBLOCK_UNIQUE_CHILD_NAME "_owtblock_unique_child_"

/*=====================================================================
                            Class Definitions
 =====================================================================*/

/*---------------------------------------------------------------------
 *                        (class) OWTHtBlock
 *
 * Base class for all other HTML manipulation classes.
 *
 * A OWTHtBlock can have children of types derived from OWTHtBlock
 *--------------------------------------------------------------------*/
class OWTHtBlock
{
protected:
    char        *m_pszName;
    char        *m_pszType;
    char        *m_pszOptions;

    char        **m_papszText;

    GBool        m_bParsed;
    int         m_numChildren;
    OWTHtBlock   **m_papoChildren;

    virtual void    Parse();

public:
    OWTHtBlock(const char *pszName=NULL, 
              const char *pszType=NULL, 
              const char *pszOptions=NULL);
    OWTHtBlock(OWTTagInfo *psTagInfo);
    virtual ~OWTHtBlock();

    virtual void    SetText(char **papszSrc, GBool bMakeCopy=TRUE);
    virtual void    SetText(char *pszNewText);
    void    SetChildText(const char *pszChildName, 
                                 char **papszSrc, GBool bMakeCopy=TRUE);
    void    SetChildText(const char *pszChildName,
                                 const char *pszText);
    virtual void    Draw(FILE *fp = NULL);

    OWTHtBlock *GetChild(const char *pszChildName);
    OWTHtBlock *SetChild(const char *pszChildName,
                                 OWTHtBlock *poNewChild);
// Those are not implemented yet.
//    virtual char  **GetText();
//    virtual char  **GetChildText(const char *pszChildName);

    virtual const char  *GetName();
    virtual void        SetName(const char *pszNewName);
    void        ExpandEnvVars(OWTCGIEnv *poCGIEnv =NULL,
                              char **papszValues = NULL);
};

/*---------------------------------------------------------------------
 *                        (class) OWTHtDoc
 *
 * Class to hold and manipulate a HTML document.
 *--------------------------------------------------------------------*/
class OWTHtDoc: public OWTHtBlock
{
protected:
    char    *m_pszContentType;

public:
    OWTHtDoc();
    virtual ~OWTHtDoc();

    virtual void    Draw(FILE *fp = NULL);

    void    SetContentType(const char *pszContentType);
    GBool   LoadFromFile(const char *pszFname);
    GBool   SaveToFile(const char *pszFname);
};

/*---------------------------------------------------------------------
 *                        (class) OWTHtTable
 *
 * Class to hold and manipulate a HTML tables.
 *--------------------------------------------------------------------*/
class OWTHtTable: public OWTHtBlock
{
protected:
    int     m_numRows;
    int     m_numCols;
    char    ***m_papapszCells;   /* [numRows][numCols] */

    int     m_numWrapColumns;
    int     m_nBorderWidth;
    GBool   m_bReversedOrientation;

    int     m_numHeaderRows;
    int     m_numHeaderCols;

    /*Text to add to the <TD>*/
    char    ***m_papapszCellAttributes;   /* [numRows][numCols] */

    void    DrawTDWithAttributes(FILE *fpOut, int nCol, int nRow,
                                 bool bReversed);

    void    DrawReversed(FILE *fpOut=NULL);

public:
    OWTHtTable();
    virtual ~OWTHtTable();

    virtual void    Draw(FILE *fp = NULL);
    virtual void    SetText(char **papszSrc, GBool bMakeCopy=TRUE);

    int     AddColumn(char **papszData);
    int     InsertColumn(int nInsertAt, char **papszData);
    int     AddRow(char **papszData);
    int     InsertRow(int nInsertAt, char **papszData);

    int     SetCell(int nCol, int nRow, const char *pszData, 
                    GBool bForceCreate = FALSE);
    int     SetCellRange(int nCol1, int nRow1, int numCols, int numRows,
                         char **papszData, GBool bForceCreate = FALSE);
    int     SetColumn(int nCol, char **papszData);
    int     SetRow(int nRow, char **papszData);

    void    SetWrapMaxCols(int numCols);
    void    SetBorderWidth(int nBorderWidth);
    void    SetReversedOrientation(GBool bReverseOrient);
    void    SetNumHeaderRows(int numHeaderRows);
    void    SetNumHeaderColumns(int numHeaderRows);

    int     GetRow();
    int     GetCol();
    int     SetCellAttribute(int nCol, int nRow, char *pszValue);
};

/*---------------------------------------------------------------------
 *                        (class) OWTTextTable
 *
 * Class to hold and manipulate a Text (tab or comma-delimited) tables.
 *
 * This class uses the exact same interface as the OWTHtTable, except 
 * that the Draw() method has been redefined.
 *--------------------------------------------------------------------*/
class OWTTextTable: public OWTHtTable
{
protected:
    char    *m_pszDelimiter;
public:
    OWTTextTable();
    virtual ~OWTTextTable();

    virtual void    Draw(FILE *fp = NULL);

    void    SetDelimiter(const char *pszDelimiter);
};


#endif /* _OWTHT_H_INCLUDED_ */
