/**********************************************************************
 * $Id: owtchartcgi.cpp,v 1.27 2003/10/14 19:09:28 daniel Exp $
 *
 * Name:     owtchartcgi.cpp
 * Project:  Open Web Toolkit
 * Language: C++
 * Purpose:  CGI interface for the GDChart library.
 * Author:   Daniel Morissette, morissette@dmsolutions.ca
 *
 **********************************************************************
 * Copyright (c) 1998-2003, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owtchartcgi.cpp,v $
 * Revision 1.27  2003/10/14 19:09:28  daniel
 * Updated PrintInfo() output
 *
 * Revision 1.26  2003/10/14 19:01:36  daniel
 * Reworked the logic to handle IMAGETYPE so that it is optional and defaults
 * are picked in the following order: PNG, GIF, JPEG.
 *
 * Revision 1.25  2003/10/14 18:18:41  daniel
 * Update version to "1.2.0 (20031014)"
 *
 * Revision 1.24  2003/10/14 17:50:32  daniel
 * Set version to 1.2.0/20021116
 *
 * Revision 1.23  2002/11/15 18:59:03  julien
 * implement Bar Labels
 *
 * Revision 1.22  2002/11/06 18:58:03  julien
 * allow one numpts for 3dhbar
 *
 * Revision 1.21  2002/08/20 13:05:16  daniel
 * Fixed improperly terminated string buffer in GifErrorHandler()
 *
 * Revision 1.20  2002/04/23 13:59:24  julien
 * Allow BAR and HBAR to have only 1 value (NumPts=1)
 *
 * Revision 1.19  2002/02/11 19:51:10  julien
 * Add Combo_Line_HLC and 3DCombo_Line_HLC type
 *
 * Revision 1.18  2002/01/21 21:30:05  julien
 * Valid if gif, png, jpeg or wbmp exist
 *
 * Revision 1.17  2002/01/14 17:35:50  julien
 * Added ImageType and JPEGQuality parameters to create GIF, PNG and JPEG
 *
 * Revision 1.16  2002/01/11 15:51:16  julien
 * Add Combo_HLC_Line and 3DCombo_HLC_Line type
 *
 * Revision 1.15  2001/12/20 18:50:57  daniel
 * Fixed PNG support in command-line mode
 *
 * Revision 1.14  2001/12/18 19:47:12  daniel
 * Added a note about PNG in usage information
 *
 * Revision 1.13  2001/12/18 19:39:06  daniel
 * Support USE_GD_PNG
 *
 * Revision 1.12  2001/11/07 22:19:18  daniel
 * Added XLabelSpacing param (value in pixels)
 *
 * Revision 1.11  2001/08/10 21:38:15  daniel
 * Added linear interpolations when VALS contains "." values.
 *
 * Revision 1.10  2001/07/30 14:46:21  assefa
 * Add function to output error in a GIF.
 *
 * Revision 1.9  2001/06/19 15:01:43  daniel
 * Added PiePctFmt parameter
 *
 * Revision 1.8  2001/04/09 16:01:23  assefa
 * Modify a bit the test to verify if the progrma is run on
 * the command line.
 *
 * Revision 1.7  2001/02/16 15:47:07  daniel
 * Added pie_requested_radius param.
 *
 * Revision 1.6  2000/11/24 18:53:12  daniel
 * Fixed error with number of expected params for ExtColors
 *
 * Revision 1.5  2000/11/22 21:49:50  daniel
 * Added YInterval option
 *
 * Revision 1.4  2000/11/20 20:00:47  daniel
 * Sync'ed gdchart2 with v0.10.1dev - added COMBO_LINE_LINE + FLOATINGBAR
 *
 * Revision 1.3  2000/11/16 21:24:13  daniel
 * Added Combo charts + accept pie charts with only 1 data value
 *
 * Revision 1.2  1999/07/27 20:12:55  daniel
 * Added PiePctType param, and fixed problem HLC charts using ExtColors
 *
 * Revision 1.1  1999/05/28 18:33:59  daniel
 * *** empty log message ***
 *
 **********************************************************************/

#define OWTCHART_VERSION  "1.2.0 (20031014)"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "cpl_error.h"
#include "cpl_string.h"
#include "owtcgien.h"

#include "gdchart.h"

extern "C" {
#include "gdfonts.h"
#include "gdfontl.h"
#include "gdfonts.h"
}

#ifdef _WIN32
#  include <io.h>     /* _setmode()   */
#  include <fcntl.h>  /* O_BINARY     */
#   define TMP_ENV   "\\temp\\cgienv.env"
#else
#   define TMP_ENV   "/tmp/cgienv.env"
#endif

static GBool gbRunAsCGI = TRUE;

void        PrintInfo();
GDCChartPtr InitChart( OWTCGIEnv *poCGIEnv);
int         SetChartOptions( GDCChartPtr psGDC, OWTCGIEnv *poCGIEnv);
void        FreeAllocatedBuffers(GDCChartPtr psGDC);


extern "C" void MyErrorHandler(CPLErr /*nErrClass*/,
                               int nErrno, const char *msg)
{
    if (gbRunAsCGI)
        fprintf(stdout, "Content-type: text/plain\n\n");
    fprintf(stdout, "ERROR %d: %s\n", nErrno, msg);

    exit(1);
}

extern "C" void GifErrorHandler(CPLErr /*nErrClass*/,
                                int nErrno, const char *msg)
{
    int nWidth = 100; /*default image size */
    int nHeight = 200;
    int nMargin =5;
    
    gdImagePtr img=NULL;
    gdFontPtr pFont = gdFontSmall;

    int nTextLength = strlen(msg);
    int nUsableWidth = 0;
    int nMaxCharsPerLine = 0;
    int nLines = 0;
    int i = 0;
    int nStart = 0;
    int nEnd = 0;
    int nLength = 0;
    char **papszLines = NULL;
    int nXPos = 0;
    int nYPos = 0;

    int nWidthTxt = 0;

    int nSpaceBewteenLines = pFont->h;

    int nWhite = 0;
    int nBlack = 0;

    FILE        *fpOut = NULL;

    
    nWidthTxt  = nTextLength * pFont->w;
    
    nUsableWidth = nWidth - (nMargin*2);
    if (nWidthTxt > nUsableWidth)
    {
        nMaxCharsPerLine =  nUsableWidth/pFont->w;
        nLines = (int) ceil ((double)nTextLength / (double)nMaxCharsPerLine);

        if (nLines > 0)
        {
            papszLines = (char **)malloc(nLines*sizeof(char *));
            for (i=0; i<nLines; i++)
            {
                papszLines[i] = (char *)malloc((nMaxCharsPerLine+1)*sizeof(char));
                papszLines[i][0] = '\0';
            }
        }
        for (i=0; i<nLines; i++)
        {
            nStart = i*nMaxCharsPerLine;
            nEnd = nStart + nMaxCharsPerLine;
            if (nStart < nTextLength)
            {
                if (nEnd > nTextLength)
                    nEnd = nTextLength;
                nLength = nEnd-nStart;

                strncpy(papszLines[i], msg+nStart, nLength);
                papszLines[i][nLength+1] = '\0';
            }
        }
    }
    else
    {
        nLines = 1;
        papszLines = (char **)malloc(nLines*sizeof(char *));
        papszLines[0] = (char *)malloc((strlen(msg)+1)*sizeof(char));
        papszLines[0] = strcpy(papszLines[0], msg);
        papszLines[0][strlen(papszLines[0])+1]='\0';
    }

    img = gdImageCreate(nWidth, nHeight);
    nWhite = gdImageColorAllocate(img, 200,200,200);  // BG color
    nBlack = gdImageColorAllocate(img, 0,0,0);        // Text color


    for (i=0; i<nLines; i++)
    {
        nYPos = (nSpaceBewteenLines) * ((i*2) +1); 
        nXPos = nSpaceBewteenLines;

        gdImageString(img, pFont, nXPos, nYPos,  
                      (unsigned char *)papszLines[i], nBlack);
    }

    if (papszLines)
    {
	for (i=0; i<nLines; i++)
	{
	    free(papszLines[i]);
	}
	free (papszLines);
    }

    fpOut = stdout;

#ifdef HAVE_GIF
    fprintf(fpOut, "Content-Type: image/gif\n\n");
    gdImageGif(img, stdout);
#endif
#ifdef HAVE_PNG
    fprintf(fpOut, "Content-Type: image/png\n\n");
    gdImagePng(img, stdout);
#endif


    gdImageDestroy(img);
}

/**********************************************************************
 *                       main()
 *
 ********************************************************************** */

int main(int argc, char *argv[]) 
{
    GDCChartPtr psGDC = NULL;
    FILE        *fpOut = NULL;
    /*-----------------------------------------------------------------
     * Create a OWTCGIEnv object that will be used to handle CGI stuff.
     * When the program exits the object will be destroyed and any 
     * resources it uses will be freed automagically.
     *----------------------------------------------------------------*/
    OWTCGIEnv oCGIEnv;

    /*-----------------------------------------------------------------
     * If -debug switch was passed, then read input from saved CGI env. file
     *----------------------------------------------------------------*/
    // oCGIEnv.DumpToFile(TMP_ENV);
#ifdef DEBUG
    if (argc >= 2 && EQUAL(argv[1], "-debug"))
        oCGIEnv.InitFromFile(TMP_ENV);
    else
#endif
    /*-----------------------------------------------------------------
     * If program does not appear to be running inside a CGI environment
     * then run as a command-line, expecting the filename of a chart
     * param. file and of an output file on the command line.
     *----------------------------------------------------------------*/
       
        if (argc >= 3 || (strlen(oCGIEnv.GetEnv("QUERY_STRING")) == 0 && 
                          strlen(oCGIEnv.GetEnv("REQUEST_METHOD")) == 0))
        {
            gbRunAsCGI = FALSE;

            if (argc == 3)
            {
                oCGIEnv.InitFromFile(argv[1]);
                if (!EQUAL(argv[2], "-"))
                {
                    fpOut = fopen(argv[2], "wb");
                    if (fpOut == NULL)
                    {
                        CPLError(CE_Failure, CPLE_FileIO,
                                 "Cannot create output file %s", argv[2]);
                    }
                }
            }
            else
            {
                printf(
        "OWTChart v%s (GDChart v%s)\n\n"
        "Usage: owtchart <param_file.txt> <out_file.gif>\n\n"
        "   <param_file.txt>: the name of the text file that contains the\n"
        "                     chart parameters\n\n" 
        "   <out_file.gif>: the name of the GIF file to create. Passing\n"
        "                   a value of '-' will send the GIF to stdout.\n\n"

        , OWTCHART_VERSION, GDCHART_VERSION);

                return 1;
            }

        }


    CPLSetErrorHandler(MyErrorHandler);

    /*-----------------------------------------------------------------
     * When URL contains "TYPE=INFO", then we print version and copyright
     * info (and do not produce any chart)
     *----------------------------------------------------------------*/
    if ( EQUAL(oCGIEnv.GetFormParam("TYPE"), "INFO") )
    {
        PrintInfo();
        return 0;
    }

    /*-----------------------------------------------------------------
     * Alloc and init the GDCChart object using the values for the 
     * compulsory parameters (TYPE, W, H, N_PTS, and VALS)
     *----------------------------------------------------------------*/
    psGDC = InitChart( &oCGIEnv );

    /*-----------------------------------------------------------------
     * Set optional chart parameters (all the other options)
     *----------------------------------------------------------------*/
    SetChartOptions( psGDC, &oCGIEnv );

    /*-----------------------------------------------------------------
     * If everything went through without problems, then output the
     * GIF image... otherwise, an error message has already been returned.
     *----------------------------------------------------------------*/
    if (psGDC && CPLGetLastErrorNo() == 0)
    {
        const char *pszReqImageType = oCGIEnv.GetFormParam("ImageType");

        /*-------------------------------------------------------------
         * Select image type, pick the first supported format if there
         * wasn't one explicitly requested.
         *-----------------------------------------------------------*/
#ifdef HAVE_PNG
        if ( pszReqImageType[0] == '\0' || EQUAL(pszReqImageType, "PNG") )
            psGDC->image_type = GDC_PNG;
        else 
#endif
#ifdef HAVE_GIF
            if ( pszReqImageType[0] == '\0' || EQUAL(pszReqImageType, "GIF") )
            {
                psGDC->image_type = GDC_GIF;
            }
            else
#endif
#ifdef HAVE_JPEG
                if ( pszReqImageType[0] == '\0' || EQUAL(pszReqImageType, "JPEG") )
                {
                    /* ---------------------------------------------
                     * The JPEGQuality is optional, default is JPEG 
                     * quality default value.
                     * --------------------------------------------*/
                    psGDC->image_type = GDC_JPEG;
                    psGDC->jpeg_quality=atoi(
                                oCGIEnv.GetFormParam("JPEGQuality"));
                    if(psGDC->jpeg_quality <= 0)
                        psGDC->jpeg_quality=-1;
                }
                else
#endif
#ifdef HAVE_WBMP
                    if ( EQUAL(pszReqImageType, "WBMP") )
                        psGDC->image_type = GDC_WBMP;
                    else 
#endif
                    {
                        CPLError(CE_Failure, CPLE_IllegalArg,
                                 "Invalid or missing IMAGETYPE parameter.");
                        return -1;
                    }
                        

        if (gbRunAsCGI)
        {
            switch( psGDC->image_type )
            {
              case GDC_PNG:    
                printf("Content-Type: image/png\n\n");
                break;
              case GDC_JPEG:  
                printf("Content-Type: image/jpeg\n\n");          
                break;
              case GDC_WBMP:  
                printf("Content-Type: image/wbmp\n\n");;         
                break;
              case GDC_GIF:
              default:        
                printf("Content-Type: image/gif\n\n");
            }

        }

        /* ----------------------------------------------------------
         * (On WIN32 we have to make sure stdout is in binary mode.)
         * ---------------------------------------------------------- */
#ifdef WIN32
       if ( fpOut == NULL &&
            _setmode( _fileno( stdout ), O_BINARY ) == -1)
            CPLError(CE_Failure, CPLE_FileIO, 
                     "Cannot set output file mode (%s)",  strerror(errno));
#endif

        GDCChartGIF(psGDC, fpOut? fpOut:stdout);
        fflush(fpOut? fpOut:stdout);
    }

    FreeAllocatedBuffers(psGDC);

    GDCChartDestroy(psGDC);

    if (fpOut)
        fclose(fpOut);

    return 0;
}





/**********************************************************************
 *                       PrintInfo()
 *
 * Return Version and Copyright information to stdout.
 ********************************************************************** */

void PrintInfo()
{
    printf("Content-Type: text/plain\n\n");

    printf("OWTChart v%s (GDChart v%s)\n\n", OWTCHART_VERSION, GDCHART_VERSION);
    printf("Copyright (c) 1999-2003, Daniel Morissette \n\n");
    printf("Uses Bruce Verderaime's GDChart library (http://www.fred.net/brv/chart/)\n");
    printf("and the GD (gd1.3) library from Thomas Boutell (http://www.boutell.com/)\n");
    printf("\n");
    printf("OWTChart home: http://www.maptools.org/owtchart/\n");
    printf("\n");
    printf("$Revision: 1.27 $  $Date: 2003/10/14 19:09:28 $ (Built %s %s)\n", __DATE__, __TIME__);
}

/**********************************************************************
 *                       InitChart()
 *
 * Allocate a new GDCChart and initialize it with the values passed for
 * the compulsory parameters. (TYPE, W, H, N_PTS, and VALS)
 *
 * If any of these parameters is missing then an error will be
 * returned.
 ********************************************************************** */

static float *gpafChartValues = NULL;

GDCChartPtr InitChart( OWTCGIEnv *poCGIEnv)
{
    GDCChartPtr psGDC = NULL;
    int         nWidth, nHeight, numPts, numSets;
    int         numHLCSets, numVolSets, numH2Sets, numExpectedVals;
    int         i;
    GDC_CHART_T eType;
    const char  *pszValue;
    char        **papszValues = NULL, **papszXLabels = NULL;
    GBool       bInterpolations = FALSE;

    /*-----------------------------------------------------------------
     * TYPE - Chart Type
     *----------------------------------------------------------------*/
    pszValue = poCGIEnv->GetFormParam("TYPE");
    if (EQUAL(pszValue, "LINE"))
        eType = GDC_LINE;
    else if (EQUAL(pszValue, "AREA"))
        eType = GDC_AREA;
    else if (EQUAL(pszValue, "BAR"))
        eType = GDC_BAR;
    else if (EQUAL(pszValue, "FLOATINGBAR"))
        eType = GDC_FLOATINGBAR;
    else if (EQUAL(pszValue, "HBAR"))
        eType = GDC_HBAR;
    else if (EQUAL(pszValue, "H2BAR"))
        eType = GDC_H2BAR;
    else if (EQUAL(pszValue, "COMBO_LINE_BAR"))
        eType = GDC_COMBO_LINE_BAR;
    else if (EQUAL(pszValue, "COMBO_HLC_BAR"))
        eType = GDC_COMBO_HLC_BAR;
    else if (EQUAL(pszValue, "COMBO_LINE_AREA"))
        eType = GDC_COMBO_LINE_AREA;
    else if (EQUAL(pszValue, "COMBO_LINE_LINE"))
        eType = GDC_COMBO_LINE_LINE;
    else if (EQUAL(pszValue, "COMBO_HLC_AREA"))
        eType = GDC_COMBO_HLC_AREA;
    else if (EQUAL(pszValue, "COMBO_HLC_LINE"))
        eType = GDC_COMBO_HLC_LINE;
    else if (EQUAL(pszValue, "COMBO_LINE_HLC"))
        eType = GDC_COMBO_LINE_HLC;
    else if (EQUAL(pszValue, "PIE"))
        eType = GDC_2DPIE;
    else if (EQUAL(pszValue, "HLC") || EQUAL(pszValue, "HILOCLOSE"))
        eType = GDC_HILOCLOSE;
    else if (EQUAL(pszValue, "3DLINE"))
        eType = GDC_3DLINE;
    else if (EQUAL(pszValue, "3DAREA"))
        eType = GDC_3DAREA;
    else if (EQUAL(pszValue, "3DBAR"))
        eType = GDC_3DBAR;
    else if (EQUAL(pszValue, "3DFLOATINGBAR"))
        eType = GDC_3DFLOATINGBAR;
    else if (EQUAL(pszValue, "3DHBAR"))
        eType = GDC_3DHBAR;
    else if (EQUAL(pszValue, "3DH2BAR"))
        eType = GDC_3DH2BAR;
    else if (EQUAL(pszValue, "3DCOMBO_LINE_BAR"))
        eType = GDC_3DCOMBO_LINE_BAR;
    else if (EQUAL(pszValue, "3DCOMBO_HLC_BAR"))
        eType = GDC_3DCOMBO_HLC_BAR;
    else if (EQUAL(pszValue, "3DCOMBO_LINE_AREA"))
        eType = GDC_3DCOMBO_LINE_AREA;
    else if (EQUAL(pszValue, "3DCOMBO_LINE_LINE"))
        eType = GDC_3DCOMBO_LINE_LINE;
    else if (EQUAL(pszValue, "3DCOMBO_HLC_AREA"))
        eType = GDC_3DCOMBO_HLC_AREA;
    else if (EQUAL(pszValue, "3DCOMBO_HLC_LINE"))
        eType = GDC_3DCOMBO_HLC_LINE;
    else if (EQUAL(pszValue, "3DCOMBO_LINE_HLC"))
        eType = GDC_3DCOMBO_LINE_HLC;
    else if (EQUAL(pszValue, "3DPIE"))
        eType = GDC_3DPIE;
    else if (EQUAL(pszValue, "3DHLC") || EQUAL(pszValue, "3DHILOCLOSE"))
        eType = GDC_3DHILOCLOSE;
    else
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "Invalid or missing TYPE parameter.");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * W, H - Optional - Chart Width and Height (Default = 300 x 200)
     *----------------------------------------------------------------*/
    nWidth = 300;  
    nHeight = 200;
    pszValue = poCGIEnv->GetFormParam("W");
    if (!EQUAL(pszValue, ""))
        nWidth = atoi( pszValue );
    pszValue = poCGIEnv->GetFormParam("H");
    if (!EQUAL(pszValue, ""))
        nHeight = atoi( pszValue );
    if (nWidth < 10 || nHeight < 10)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "Invalid W (width) and/or H (height) parameter.");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * NumSets - Optional, default is 1 
     *----------------------------------------------------------------*/
    pszValue = poCGIEnv->GetFormParam("NumSets");
    if (EQUAL(pszValue, ""))
        numSets = 1;
    else if ( (numSets = atoi(pszValue)) < 1)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "Invalid NumSets parameter (must be greather than 0).");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * N_PTS - Num. of values along X axis (in 1 set) 
     *----------------------------------------------------------------*/
    numPts = atoi(poCGIEnv->GetFormParam("NumPts"));
    if (numPts < 1 || 
        (numPts < 2 && !(eType == GDC_2DPIE || eType == GDC_3DPIE || 
                         eType == GDC_HBAR || eType == GDC_H2BAR || 
                         eType == GDC_3DHBAR || eType == GDC_3DH2BAR || 
                         eType == GDC_BAR || eType == GDC_3DBAR)))
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "Invalid NumPts parameter (must be at least 2 for XY charts "
                 "and 1 for Pie charts).");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * VALS - Values to chart 
     * Note: we have to store the values in a static buffer that will
     *       be freed only after the chart is drawn.
     *----------------------------------------------------------------*/

    /*-----------------------------------------------------------------
     * Start by validating the user's input
     *----------------------------------------------------------------*/
    pszValue = poCGIEnv->GetFormParam("VALS");
    if (EQUAL(pszValue, ""))
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "Missing VALS parameter.");
        return NULL;
    }
    numHLCSets = ( eType == GDC_COMBO_HLC_BAR   ||
                   eType == GDC_COMBO_HLC_AREA  ||
                   eType == GDC_COMBO_HLC_LINE  ||
                   eType == GDC_COMBO_LINE_HLC  ||
                   eType == GDC_3DCOMBO_HLC_BAR ||
                   eType == GDC_3DCOMBO_HLC_AREA||
                   eType == GDC_3DCOMBO_HLC_LINE||
                   eType == GDC_3DCOMBO_LINE_HLC||
                   eType == GDC_3DHILOCLOSE     ||
                   eType == GDC_HILOCLOSE )? numSets: 0;
    numVolSets = (eType == GDC_COMBO_HLC_BAR   || /* aka: combo */
                  eType == GDC_COMBO_HLC_AREA  ||
                  eType == GDC_COMBO_HLC_LINE  ||
                  eType == GDC_COMBO_LINE_HLC  ||
                  eType == GDC_COMBO_LINE_BAR  ||
                  eType == GDC_COMBO_LINE_AREA ||
                  eType == GDC_COMBO_LINE_LINE ||
                  eType == GDC_3DCOMBO_HLC_BAR ||
                  eType == GDC_3DCOMBO_HLC_AREA||
                  eType == GDC_3DCOMBO_HLC_LINE||
                  eType == GDC_3DCOMBO_LINE_HLC||
                  eType == GDC_3DCOMBO_LINE_BAR||
                  eType == GDC_3DCOMBO_LINE_AREA||
                  eType == GDC_3DCOMBO_LINE_LINE ) ? numSets: 0;
    numH2Sets = ( eType == GDC_H2BAR ||
                  eType == GDC_3DH2BAR ) ? numSets: 0;
    if (numHLCSets > 0)
        numExpectedVals = (numHLCSets*3 + numVolSets) * numPts;
    else if (numH2Sets > 0)
        numExpectedVals = (numH2Sets*2 + numVolSets) * numPts;
    else
        numExpectedVals = (numSets + numVolSets) * numPts;

    papszValues = CSLTokenizeStringComplex(pszValue, ";_!", TRUE, TRUE);
    if (CSLCount(papszValues) != numExpectedVals)
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "VALS parameter: Invalid number of values "
                 "(got %d, expected %d).", 
                 CSLCount(papszValues), numExpectedVals);
        CSLDestroy(papszValues);
        return NULL;
    }

    /*-----------------------------------------------------------------
     * Everything seems OK ... copy the values in a float array that
     * will be acceptable for the GDChart lib later.
     *----------------------------------------------------------------*/
    double dTotal = 0.0;
    gpafChartValues = (float*)CPLCalloc(numExpectedVals, sizeof(float));
    for(i=0; i<numExpectedVals; i++)
    {
        if (EQUAL(papszValues[i], ""))
            gpafChartValues[i] = GDC_NOVALUE;  /* Missing Values */
        else if (EQUAL(papszValues[i], "."))
        {
            gpafChartValues[i] = GDC_INTERP_VALUE;  /* Interpolated Values */
            bInterpolations = TRUE; /* We'll enable interpolations in psGDC */
        }
        else
        {
            gpafChartValues[i] = atof(papszValues[i]);
            dTotal += gpafChartValues[i];
        }
    }

    CSLDestroy(papszValues);

    /*-----------------------------------------------------------------
     * For PIE charts, check that the total of the values is > 0,
     * otherwise we can run into infinite loops.
     *----------------------------------------------------------------*/
    if (dTotal <= 0.0 && (eType == GDC_2DPIE || eType == GDC_3DPIE) )
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
               "VALS parameter: Cannot produce a pie chart with null values");
        return NULL;
    }

    /*-----------------------------------------------------------------
     * XLabels - Optional, if it is not provided then we'll just build
     *           numeric labels (1, 2, 3, 4, ... ), except for PIE
     *           charts, for which we set empty labels by default. ("")
     * Note: papszXLabels has to remain valid until the chart is
     *       finished redrawing... it will be stored in the psGDC 
     *       struct and will be freed by FreeAllocatedBuffers()
     *----------------------------------------------------------------*/
    pszValue = poCGIEnv->GetFormParam("XLabels");
    OWTDecodeHTML2ISOLatin1((char*)pszValue);
    if (! EQUAL(pszValue, "") &&
        ( (papszXLabels = CSLTokenizeStringComplex(pszValue, ";", 
                                                   TRUE, TRUE)   ) == NULL ||
          CSLCount(papszXLabels) != numPts )  )
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "XLabels parameter: Invalid number of labels "
                 "(got %d, expected %d).", 
                 CSLCount(papszXLabels), numPts);
        CSLDestroy(papszXLabels);
        return NULL;        
    }
    else if (eType == GDC_2DPIE || eType == GDC_3DPIE)
    {
        for(i=1; i <= numPts; i++)
            papszXLabels = CSLAddString(papszXLabels, "");
    }
    else 
    {
        for(i=1; i <= numPts; i++)
            papszXLabels = CSLAppendPrintf(papszXLabels, "%d", i);
    }


    /*-----------------------------------------------------------------
     * Create the GDCChart structure and initialize it.
     *----------------------------------------------------------------*/
    if ((psGDC = GDCChartAlloc(eType, nWidth, nHeight)) == NULL ||
        GDCChartSetValues(psGDC, numPts, papszXLabels, 
                          numSets, gpafChartValues) != 0  )
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "GDCChartPtr initialization failed! "
                 "Make sure input parameters are valid");
        return NULL;        
    }

    psGDC->interpolations = bInterpolations;

    return psGDC;
}


/*-----------------------------------------------------------------
 * Define some macros and functions to simplify setting of options
 *----------------------------------------------------------------*/

#define SET_COLOR_OPTION(dst_var, param_name)  \
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam( param_name )), "") )  \
    {                                                                     \
        dst_var = (GDC_RGB_T)strtoul((char*)pszValue, NULL, 16);          \
    }

#define SET_STRING_OPTION(dst_var, param_name)  \
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam( param_name )), "") )  \
    {                                                                     \
        OWTDecodeHTML2ISOLatin1((char*)pszValue);                         \
        dst_var = (char*)pszValue;                                        \
    }

#define SET_INT_OPTION(type, dst_var, param_name)  \
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam( param_name )), "") )  \
    {                                                                     \
        dst_var = (type) atoi((char*)pszValue);                           \
    }

#define SET_FLOAT_OPTION( dst_var, param_name)  \
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam( param_name )), "") )  \
    {                                                                     \
        dst_var = atof((char*)pszValue);                           \
    }

#define SET_FONT_OPTION( dst_var, param_name)  \
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam( param_name )), "") )  \
    {                                                                     \
        switch(toupper(pszValue[0])) {                  \
          case 'T': dst_var = GDC_TINY; break;          \
          case 'S': dst_var = GDC_SMALL; break;         \
          case 'M': dst_var = GDC_MEDBOLD; break;       \
          case 'L': dst_var = GDC_LARGE; break;         \
          case 'G': dst_var = GDC_GIANT; break;         \
          default:                                      \
            CPLError(CE_Failure, CPLE_IllegalArg,       \
                     "Invalid %s parameter value (%s)", \
                     param_name, pszValue);             \
            return -1;                                  \
         }                                              \
    }

/**********************************************************************
 *                       _SetColorListtOption()
 *
 * Read a string that contains a list of hex. color values and return
 * a newly allocated array with the values.
 ********************************************************************** */
GDC_RGB_T *_SetColorListOption(const char *pszParamName,
                               const char *pszValue, int numExpectedVals)
{
    char        **papszColors;
    GDC_RGB_T   *paColors = NULL;
    int         i;

    if (EQUAL(pszValue, ""))
        return NULL;  /* Nothing to do */

    papszColors = CSLTokenizeStringComplex(pszValue, ";_!", TRUE, TRUE);
    if (CSLCount(papszColors) == numExpectedVals)
    {
        paColors = (GDC_RGB_T*)CPLCalloc(numExpectedVals, sizeof(GDC_RGB_T));
        for(i=0; i<numExpectedVals; i++)
            paColors[i] = (GDC_RGB_T)strtoul(papszColors[i], NULL, 16);
    }
    else
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "%s parameter: Invalid number of colors "
                 "(got %d, expected %d).", 
                 pszParamName, CSLCount(papszColors), numExpectedVals);
    }

    CSLDestroy(papszColors);
    return paColors;
}

/**********************************************************************
 *                       SetChartOptions()
 *
 * Set the optional chart parameters.
 ********************************************************************** */

int SetChartOptions( GDCChartPtr psGDC, OWTCGIEnv *poCGIEnv)
{
    const char  *pszValue;

    if (psGDC == NULL || psGDC->num_points*psGDC->num_sets == 0 )
        return -1;

    /*-----------------------------------------------------------------
     * First override some of the defaults of the GDChart library
     *----------------------------------------------------------------*/
    psGDC->BGColor = 0xFFFFFF;          /* White BG */
    psGDC->LineColor = 0x000000;        /* Black axis */
    psGDC->PlotColor = 0x0000FF;        /* Plot values in blue */
    psGDC->XTitleColor = 0x000000;      /* X Axis Title (Def = PlotColor) */
    psGDC->YTitleColor = 0x000000;      /* X Axis Title (Def = PlotColor) */
    psGDC->ylabel_density = 50;         /* % */
    psGDC->ylabel_fmt = "%g";           /* Default was fractions */
    psGDC->ylabel2_fmt = "%g";          /* Default was fractions */
    psGDC->pie_EdgeColor = 0x000000;    /* Black Edges */
    psGDC->pie_percent_labels = GDCPIE_PCT_RIGHT;
    psGDC->pie_label_line = TRUE;
    psGDC->pie_label_dist = 15;

    /*-----------------------------------------------------------------
     * Default to have diff. color for each slice of a pie chart
     *----------------------------------------------------------------*/
    if (psGDC->type == GDC_2DPIE || psGDC->type == GDC_3DPIE)
    {
        int         i;
        psGDC->pie_Color = (GDC_RGB_T*)CPLCalloc(psGDC->num_points,
                                                 sizeof(GDC_RGB_T));
        for(i=0; i<psGDC->num_points; i++)
        {
            /* This method should allow us to generate 32 different colors.
             */
            psGDC->pie_Color[i] = ( 0x010000*200*((i+1)&1)/1 +
                                    0x000100*200*((i+1)&2)/2 +
                                    0x000001*200*((i+1)&4)/4 );
        }
    }

    /*-----------------------------------------------------------------
     * TITLEs
     *----------------------------------------------------------------*/
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("TITLE")), "") )
    {
        /* Replace semicolons with '\n' chars.  buffer will be freed 
         * by FreeAllocatedBuffers()
         */
        char *pChar;
        OWTDecodeHTML2ISOLatin1((char*)pszValue);
        pChar = psGDC->title = CPLStrdup(pszValue);
        while( (pChar = strchr(pChar, ';')) != NULL)
            *pChar = '\n';
    }

    SET_STRING_OPTION(psGDC->xtitle,    "XTitle")
    SET_STRING_OPTION(psGDC->ytitle,    "YTitle")
    SET_STRING_OPTION(psGDC->ytitle2,   "YTitle2")

    /* There is a bug with I.E. 4.0 and Navigator 3.0, related to the
     * use of "&LTitle=value" inside a IMG SRC URL... the "&LT" is seen
     * as a "<" character by these browsers and this messes things up.
     * We will continue supporting the use of "LTitle", but the use of 
     * "L_Title" is preferred.
     */
    SET_STRING_OPTION(psGDC->ltitle,    "LTitle")
    SET_STRING_OPTION(psGDC->rtitle,    "RTitle")
    SET_STRING_OPTION(psGDC->ltitle,    "L_Title")
    SET_STRING_OPTION(psGDC->rtitle,    "R_Title")

    /* Bar labels */
    SET_INT_OPTION(char, psGDC->bar_labels,     "BarLabels")
    SET_STRING_OPTION(psGDC->bar_label_fmt,  "BarLabelFmt")

    /*-----------------------------------------------------------------
     * COLORs
     *----------------------------------------------------------------*/
    SET_COLOR_OPTION(psGDC->BGColor,    "BGColor")
    SET_COLOR_OPTION(psGDC->GridColor,  "GRIDColor")
    SET_COLOR_OPTION(psGDC->LineColor,  "LINEColor")
    SET_COLOR_OPTION(psGDC->PlotColor,  "PLOTColor")
    SET_COLOR_OPTION(psGDC->VolColor,   "VOLColor")
    SET_COLOR_OPTION(psGDC->TitleColor, "TITLEColor")
    SET_COLOR_OPTION(psGDC->XTitleColor, "XTITLEColor")
    SET_COLOR_OPTION(psGDC->YTitleColor, "YTITLEColor")
    SET_COLOR_OPTION(psGDC->YTitle2Color,"YTITLE2Color")
    SET_COLOR_OPTION(psGDC->XLabelColor, "XLABELColor")
    SET_COLOR_OPTION(psGDC->YLabelColor, "YLABELColor")
    SET_COLOR_OPTION(psGDC->YLabel2Color,"YLABEL2Color")
    SET_STRING_OPTION(psGDC->ylabel_fmt, "YLABELFmt")
    SET_STRING_OPTION(psGDC->ylabel2_fmt,"YLABEL2Fmt")

    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("SetColors")), "") )
    {
        CPLFree(psGDC->SetColor);
        psGDC->SetColor = _SetColorListOption("SetColors", pszValue,
                                              psGDC->num_sets);
    }

    /*-----------------------------------------------------------------
     * ExtColors - Set the color for each chart value.
     * The actual member of the structure the value is stored in differs for
     * XY and PIE charts.
     *----------------------------------------------------------------*/
    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("ExtColors")), "") )
    {
        if (psGDC->type == GDC_2DPIE || psGDC->type == GDC_3DPIE)
        {
            CPLFree(psGDC->pie_Color);
            psGDC->pie_Color = _SetColorListOption("ExtColors", pszValue,
                                                   psGDC->num_points);
        }
        else
        {
            CPLFree(psGDC->ExtColor);
            psGDC->ExtColor = _SetColorListOption("ExtColors", pszValue,
                                           psGDC->num_sets*psGDC->num_points);
        }
    }

    /*-----------------------------------------------------------------
     * FONTs
     *----------------------------------------------------------------*/
    SET_FONT_OPTION(psGDC->title_size,    "TitleFont")
    SET_FONT_OPTION(psGDC->xtitle_size,   "XTitleFont")
    SET_FONT_OPTION(psGDC->ytitle_size,   "YTitleFont")
    SET_FONT_OPTION(psGDC->xaxisfont_size,"XLabelFont")
    SET_FONT_OPTION(psGDC->yaxisfont_size,"YLabelFont")
    SET_FONT_OPTION(psGDC->pie_label_size,"PieLabelFont")


    /*-----------------------------------------------------------------
     * LineStyles - Array of num_sets strings containing each a line
     *              style definition in which a space means a transparent
     *              pixel, and anything else means to draw the pixel.
     *              The chars ";!_" are used as delimiters in this param.
     * Note: papszStyles has to remain valid until the chart is
     *       finished redrawing... it will be stored in the psGDC 
     *       struct and will be freed by FreeAllocatedBuffers()
     *----------------------------------------------------------------*/
    char **papszStyles = NULL;
    pszValue = poCGIEnv->GetFormParam("LineStyles");
    if (! EQUAL(pszValue, "") &&
        ( (papszStyles = CSLTokenizeStringComplex(pszValue, ";!_", 
                                                   TRUE, TRUE)   ) == NULL ||
          CSLCount(papszStyles) != psGDC->num_sets )  )
    {
        CPLError(CE_Failure, CPLE_IllegalArg,
                 "LineStyles parameter: Invalid number of styles "
                 "(got %d, expected %d).", 
                 CSLCount(papszStyles), psGDC->num_sets);
        CSLDestroy(papszStyles);
        return -1;
    }
    psGDC->LineStyles = papszStyles;

    /*-----------------------------------------------------------------
     * Others
     *----------------------------------------------------------------*/
    SET_INT_OPTION(char, psGDC->transparent_bg,"TranspBG")
    SET_INT_OPTION(GDC_TICK_T, psGDC->grid,    "Grid")
    SET_INT_OPTION(char, psGDC->xaxis,        "XAxis")
    SET_INT_OPTION(char, psGDC->yaxis,        "YAxis")
    SET_INT_OPTION(char, psGDC->yaxis2,       "YAxis2")
    SET_INT_OPTION(char, psGDC->yval_style,   "YValStyle")
    SET_INT_OPTION(char, psGDC->xlabel_spacing,"XLabelSpacing")

    SET_INT_OPTION(int, psGDC->_3d_depth,    "3DDepth")
    SET_INT_OPTION(int, psGDC->_3d_angle,    "3DAngle")
    SET_INT_OPTION(int, psGDC->bar_width,    "BarWidth")

    SET_INT_OPTION(GDC_HLC_STYLE_T, psGDC->HLC_style, "HLCStyle")
    SET_INT_OPTION(int, psGDC->HLC_cap_width, "HLCCapWidth")

    SET_FLOAT_OPTION(psGDC->requested_ymin,   "YMin")
    SET_FLOAT_OPTION(psGDC->requested_ymax,   "YMax")
    SET_FLOAT_OPTION(psGDC->requested_yinterval,"YInterval")
    SET_FLOAT_OPTION(psGDC->requested_ymin2,  "YMin2")
    SET_FLOAT_OPTION(psGDC->requested_ymax2,  "YMax2")

    SET_FLOAT_OPTION(psGDC->pie_requested_radius, "PieRadius")

    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("StackType")), "") )
    {
        if (EQUAL(pszValue, "DEPTH")) 
            psGDC->stack_type = GDC_STACK_DEPTH;
        else if (EQUAL(pszValue, "SUM")) 
            psGDC->stack_type = GDC_STACK_SUM;
        else if (EQUAL(pszValue, "BESIDE")) 
            psGDC->stack_type = GDC_STACK_BESIDE;
        else if (EQUAL(pszValue, "LAYER")) 
            psGDC->stack_type = GDC_STACK_LAYER;
        else
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "Invalid StackType parameter value (%s)", pszValue);
            return -1;        
        }
    }

    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("YScaleType")), "") )
    {
        if (EQUALN(pszValue, "LIN", 3)) 
            psGDC->yscale_type = GDC_SCALE_LINEAR;
        else if (EQUALN(pszValue, "LOG", 3)) 
            psGDC->yscale_type = GDC_SCALE_LOG;
        else
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "Invalid YScaleType parameter value (%s)", pszValue);
            return -1;        
        }
    }

    SET_STRING_OPTION(psGDC->pie_percent_fmt,"PiePctFmt")

    if ( !EQUAL((pszValue = poCGIEnv->GetFormParam("PiePctType")), "") )
    {
        if (EQUAL(pszValue, "NONE")) 
            psGDC->pie_percent_labels = GDCPIE_PCT_NONE;
        else if (EQUAL(pszValue, "ABOVE")) 
            psGDC->pie_percent_labels = GDCPIE_PCT_ABOVE;
        else if (EQUAL(pszValue, "BELOW")) 
            psGDC->pie_percent_labels = GDCPIE_PCT_BELOW;
        else if (EQUAL(pszValue, "RIGHT")) 
            psGDC->pie_percent_labels = GDCPIE_PCT_RIGHT;
        else if (EQUAL(pszValue, "LEFT")) 
            psGDC->pie_percent_labels = GDCPIE_PCT_LEFT;
        else
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "Invalid PiePctType parameter value (%s)", pszValue);
            return -1;        
        }
    }

    return 0;
}

/**********************************************************************
 *                       FreeAllocatedBuffers()
 *
 * Release the memory used by the buffers that are not maintained
 * by the GDChart library.
 ********************************************************************** */

void FreeAllocatedBuffers( GDCChartPtr psGDC )
{
    if (psGDC == NULL)
        return;

    CPLFree(gpafChartValues);
    gpafChartValues = NULL;

    CSLDestroy(psGDC->xlbl);
    psGDC->xlbl = NULL;

    CPLFree(psGDC->SetColor);
    psGDC->SetColor = NULL;

    CPLFree(psGDC->ExtColor);
    psGDC->ExtColor = NULL;

    CPLFree(psGDC->pie_Color);
    psGDC->pie_Color = NULL;

    CPLFree(psGDC->title);
    psGDC->title = NULL;

}














