/**********************************************************************
 * $Id: owthtcomblock.cpp,v 1.1.1.1 1999/06/04 19:37:36 daniel Exp $
 *
 * Name:     owthtcomblock.cpp
 * Project:  OWT HTML manipulation library - COM Interface
 * Purpose:  Implementation of the OWTHtComBlock wrapper class
 * Author:   Daniel Morissette, danmo@videotron.ca
 *
 **********************************************************************
 * Copyright (c) 1999, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************
 *
 * $Log: owthtcomblock.cpp,v $
 * Revision 1.1.1.1  1999/06/04 19:37:36  daniel
 * First kind of complete version
 *
 **********************************************************************/



/* The OWTHtComBlock class encapsulates a OWTCHt* component and allows
 * to set a OWTComponent as a child of an OWTHt* object.
 *
 * For instance, you could have a document created in VB that contains 
 * a table that you want to create in VB and store in the document
 * before you reDraw() it.  The Table is a COM object created in VB, 
 * but since most of the OWTHt* classes are not COM-aware, we encapsulate
 * the Table component inside a OWTHtComBlock wrapper object that can be
 * stored as a child in a OWTHtDoc:
 *
      OWTCHtDoc
        |
        +- OWTHtDoc *m_pDoc
              |
              +-- Child named "TTT", is a table:
                  OWTHtComBlock *pChild
                       |
                       +-- OWTHtBlock *m_pBlock = OWTHtTable in reality

 */
 

#include <objbase.h>
#include <stdio.h>

#include "Iface.h"
#include "ic_Util.h"
#include "ic_CUnknown.h"
#include "ic_CFactory.h" // Needed for module handle
#include "owtcom_cmpnt.h"
#include "oledb_sup.h"
#include "owthtcomblock.h"

static inline void trace(const char* msg)
	{ Util::Trace("OWTHtComBlock", msg, S_OK) ;}
static inline void trace(const char* msg, HRESULT hr)
	{ Util::Trace("OWTHtComBlock", msg, hr) ;}


/*=====================================================================
                        OWTHtComBlock class.
 =====================================================================*/

/**********************************************************************
 *                       OWTHtComBlock::OWTHtComBlock()
 *
 **********************************************************************/
OWTHtComBlock::OWTHtComBlock(IUnknown *pIUnknown):
                     OWTHtBlock(NULL, NULL, NULL)
{
    HRESULT hr;
    m_pIBlock = NULL;

    trace("Constructing OWTHtComBlock component wrapper");
    CPLErrorReset();

    if (pIUnknown)
    {
        /*-------------------------------------------------------------
         * Look for the first interface that will satisfy our needs
         *
         * The call to QueryInterface() automatically increments the
         * reference count for the interface... we'll call Release()
         * in the destructor.
         *
         * Using IDispatch allows us to work with the default interface
         * of any object, even VB components!!!
         *------------------------------------------------------------*/
//        hr = pIUnknown->QueryInterface(IID_IOWTCHtBlock, 
//                                       (void**)&m_pIBlock);
        hr = pIUnknown->QueryInterface(IID_IDispatch, 
                                       (void**)&m_pIBlock);

//      if (FAILED(hr))
//          hr = pIUnknown->QueryInterface(IID_IOWTCHtTable,
//                                           (void**)&m_pIBlock);
        if (FAILED(hr))
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "OWTHtComBlock(): IDispatch not available.");
            trace(CPLGetLastErrorMsg(), hr);
            return;
        }

        /*-------------------------------------------------------------
         * Fetch the DISPID for the functions that interest us
         *------------------------------------------------------------*/
        OLECHAR *name;
        name = L"SetName";
        hr = m_pIBlock->GetIDsOfNames(IID_NULL,      // Must be IID_NULL
                                    &name,           // Function name
                                    1,               // Nbr of Names
                                    0x00,            // LCID
                                    &m_dispidSetName);
        if (FAILED(hr))
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "OWTHtComBlock: SetName method not available.");
            trace(CPLGetLastErrorMsg(), hr);
            return;
        }
                          
        name = L"GetName";
        hr = m_pIBlock->GetIDsOfNames(IID_NULL,      // Must be IID_NULL
                                    &name,           // Function name
                                    1,               // Nbr of Names
                                    0x00,            // LCID
                                    &m_dispidGetName);
        if (FAILED(hr))
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "OWTHtComBlock: GetName method not available.");
            trace(CPLGetLastErrorMsg(), hr);
            return;
        }

        name = L"SetText";
        hr = m_pIBlock->GetIDsOfNames(IID_NULL,      // Must be IID_NULL
                                    &name,           // Function name
                                    1,               // Nbr of Names
                                    0x00,            // LCID
                                    &m_dispidSetText);
        if (FAILED(hr))
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "OWTHtComBlock: SetText method not available.");
            trace(CPLGetLastErrorMsg(), hr);
            return;
        }

        name = L"Draw";
        hr = m_pIBlock->GetIDsOfNames(IID_NULL,      // Must be IID_NULL
                                    &name,           // Function name
                                    1,               // Nbr of Names
                                    0x00,            // LCID
                                    &m_dispidDraw);
        if (FAILED(hr))
        {
            CPLError(CE_Failure, CPLE_IllegalArg,
                     "OWTHtComBlock: Draw method not available.");
            trace(CPLGetLastErrorMsg(), hr);
            return;
        }

    }
}

/**********************************************************************
 *                       OWTHtComBlock::~OWTHtComBlock()
 *
 **********************************************************************/
OWTHtComBlock::~OWTHtComBlock()
{
    trace("Destroy self.");

    if (m_pIBlock)
        m_pIBlock->Release();
}


/**********************************************************************
 *                       OWTHtComBlock::GetName()
 *
 * Return this block's name, or "" if no name has been set for it.
 * The returned value is should not be modified nor freed by the caller.
 **********************************************************************/
const char *OWTHtComBlock::GetName()
{
    char        *pszName=NULL;
    DISPPARAMS  sDispParams = {NULL, NULL, 0, 0};
    VARIANT     varResult;
    HRESULT     hr;

    trace("GetName()");
    if (m_pIBlock == NULL)
    {
        trace("OWTHtComBlock: m_pIBlock is NULL!!!");
        return "";
    }

    hr = m_pIBlock->Invoke(m_dispidGetName,     // DISPID
                           IID_NULL,
                           0x00,                // LCID
                           DISPATCH_METHOD,
                           &sDispParams,        // Method args.
                           &varResult,          // Results
                           NULL,                // Exception
                           NULL);               // Arg error
    if (FAILED(hr) || varResult.vt != VT_BSTR)
    {
        trace("OWTHtComBlock: Invoke GetName() failed!", hr);
        CPLError(CE_Failure, CPLE_AppDefined,
                 "OWTHtComBlock: Invoke GetName() failed!", hr);
        return "";
    }

    /*-----------------------------------------------------------------
     * OK, we got the name from the component... use the OWTHtBlock
     * Get/SetName() methods to do the rest of the work.
     *----------------------------------------------------------------*/
    UnicodeToAnsi(varResult.bstrVal, &pszName);
    OWTHtBlock::SetName(pszName);
    CoTaskMemFree(pszName);

    SysFreeString(varResult.bstrVal);

    return OWTHtBlock::GetName();
}


/**********************************************************************
 *                       OWTHtComBlock::SetName()
 *
 * Set a new value for this object's name.
 **********************************************************************/
void OWTHtComBlock::SetName(const char *pszNewName)
{
    char        *pszName=NULL;
    DISPPARAMS  sDispParams;
    VARIANTARG  varArg;
    HRESULT     hr;

    trace("SetName()");

    if (m_pIBlock == NULL)
    {
        trace("OWTHtComBlock: m_pIBlock is NULL!!!");
        return;
    }

    /*-----------------------------------------------------------------
     * Prepare parameters and call the method
     *----------------------------------------------------------------*/
    VariantInit(&varArg);
    varArg.vt = VT_BSTR;
    AnsiToBSTR(pszNewName, &varArg.bstrVal);

    sDispParams.cArgs = 1;      // 1 argument
    sDispParams.rgvarg = &varArg;
    sDispParams.cNamedArgs = 0;
    sDispParams.rgdispidNamedArgs = NULL;

    hr = m_pIBlock->Invoke(m_dispidSetName,     // DISPID
                           IID_NULL,
                           0x00,                // LCID
                           DISPATCH_METHOD,
                           &sDispParams,        // Method args.
                           NULL,                // Results
                           NULL,                // Exception
                           NULL);               // Arg error
    if (FAILED(hr))
    {
        trace("OWTHtComBlock: Invoke SetName() failed!", hr);
        CPLError(CE_Failure, CPLE_AppDefined,
                 "OWTHtComBlock: Invoke SetName() failed!", hr);
        return;
    }

    /*-----------------------------------------------------------------
     * Update the members of the OWTHtBlock as well and release memory
     *----------------------------------------------------------------*/
    OWTHtBlock::SetName(pszNewName);

    SysFreeString(varArg.bstrVal);

}



/**********************************************************************
 *                       OWTHtComBlock::SetText()
 *
 * Set the text contents of the current object.  This will replace
 * any previous contents, and reset the object to a not-parsed state.
 *
 * SetText() defaults to allocate its own copy of papszSrc, but if
 * bMakeCopy is passed as FALSE, then the current object becomes the
 * owner of the StringList... and it will take care of freeing it when
 * time comes.
 **********************************************************************/
void OWTHtComBlock::SetText(char **papszSrc, GBool bMakeCopy /* =TRUE */)
{
    //__TODO__
    trace("SetText() ... not implemented yet.");

}

/**********************************************************************
 *                       OWTHtComBlock::SetText()
 *
 * Same as above, but takes a single string as input.
 **********************************************************************/

void OWTHtComBlock::SetText(char *pszNewText)
{
    char *apszText[2];

    /* Create a temporary StringList and call the other SetText()
     */
    apszText[0] = pszNewText;
    apszText[1] = NULL;

    SetText(apszText);
}

/**********************************************************************
 *                       OWTHtComBlock::Draw()
 *
 * Print the contents of the block to the specified file.
 *
 * Default output file is stdout.
 **********************************************************************/
void OWTHtComBlock::Draw(FILE *fpOut /*=NULL*/)
{
    int nFile;
    DISPPARAMS  sDispParams;
    VARIANTARG  varArg;
    HRESULT     hr;

    trace("Draw()");

    /*-----------------------------------------------------------------
     * Convert fpOut to an int file handle to pass to the component's 
     * Draw() method... this will probably work only for inproc servers.
     *----------------------------------------------------------------*/
    if (fpOut == NULL)
        fpOut = stdout;

    nFile = fileno(fpOut);
    char szBuf[100];
    sprintf(szBuf, "Output file handle is %d", nFile);
    trace(szBuf);

    /*-----------------------------------------------------------------
     * Prepare parameters and call the method
     *----------------------------------------------------------------*/
    VariantInit(&varArg);
    varArg.vt = VT_I4;
    varArg.lVal = nFile;

    sDispParams.cArgs = 1;      // 1 argument
    sDispParams.rgvarg = &varArg;
    sDispParams.cNamedArgs = 0;
    sDispParams.rgdispidNamedArgs = NULL;

    hr = m_pIBlock->Invoke(m_dispidDraw,        // DISPID
                           IID_NULL,
                           0x00,                // LCID
                           DISPATCH_METHOD,
                           &sDispParams,        // Method args.
                           NULL,                // Results
                           NULL,                // Exception
                           NULL);               // Arg error
    if (FAILED(hr))
    {
        trace("OWTHtComBlock: Invoke Draw() failed!", hr);
        CPLError(CE_Failure, CPLE_AppDefined,
                 "OWTHtComBlock: Invoke Draw() failed!", hr);
        return;
    }
}


